"""Definition of middleware that will generate BotX API errors depending from flag."""
from typing import Tuple, Type

from pydantic import BaseModel
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Match

from botx.clients.methods.base import APIErrorResponse, BotXMethod
from botx.testing.botx_mock.asgi.responses import PydanticResponse


def _fill_request_scope(request: Request) -> None:
    routes = request.app.router.routes
    for route in routes:
        match, scope = route.matches(request)
        if match == Match.FULL:
            request.scope = {**request.scope, **scope}


def _get_error_from_request(
    request: Request,
) -> Tuple[Type[BotXMethod], Tuple[int, BaseModel]]:
    _fill_request_scope(request)
    endpoint = request.scope["endpoint"]
    method = endpoint.method
    return method, request.app.state.errors.get(method)


def should_generate_error_response(request: Request) -> bool:
    """Check if mocked API should generate error response.

    Arguments:
        request: request from Starlette route that contains application with required
            state.

    Returns:
        Result of check.
    """
    _, status_and_error_to_raise = _get_error_from_request(request)
    return bool(status_and_error_to_raise)


def generate_error_response(request: Request) -> Response:
    """Generate error response for mocked BotX API.

    Arguments:
        request: request from Starlette route that contains application with required
            state.

    Returns:
        Generated response.
    """
    method, response_info = _get_error_from_request(request)
    status_code, error_data = response_info

    return PydanticResponse(
        APIErrorResponse[BaseModel](
            errors=["error from mock"], reason="asked_for_error", error_data=error_data,
        ),
        status_code=status_code,
    )


class ErrorMiddleware(BaseHTTPMiddleware):
    """Middleware that will generate error response."""

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint,
    ) -> Response:
        """Generate error response for API call or pass request to mocked endpoint.

        Arguments:
            request: request that should be handled.
            call_next: next executor for mock.

        Returns:
              Mocked response.
        """
        if should_generate_error_response(request):
            return generate_error_response(request)

        return await call_next(request)



"""Pydantic models for mentions."""

from typing import Optional, Union
from uuid import UUID, uuid4

from pydantic import BaseModel, validator

from botx.models.enums import MentionTypes


class UserMention(BaseModel):
    """Mention for single user in chat or by `user_huid`."""

    #: huid of user that will be mentioned.
    user_huid: UUID

    #: name that will be used instead of default user name.
    name: Optional[str] = None


class ChatMention(BaseModel):
    """Mention chat in message by `group_chat_id`."""

    #: id of chat that will be mentioned.
    group_chat_id: UUID

    #: name that will be used instead of default chat name.
    name: Optional[str] = None


class Mention(BaseModel):
    """Mention that is used in bot in messages."""

    #: unique id of mention.
    mention_id: Optional[UUID] = None

    #: mention type.
    mention_data: Union[ChatMention, UserMention]

    #: payload with data about mention.
    mention_type: MentionTypes = MentionTypes.user

    @validator("mention_id", pre=True, always=True)
    def generate_mention_id(cls, mention_id: Optional[UUID]) -> UUID:  # noqa: N805
        """Verify that `mention_id` will be in mention.

        Arguments:
            cls: this class.
            mention_id: id that should present or new UUID4 will be generated.

        Returns:
             Mention ID.
        """
        return mention_id or uuid4()

    @validator("mention_type", pre=True, always=True)
    def check_that_type_matches_data(
        cls, mention_type: MentionTypes, values: dict,  # noqa: N805, WPS110
    ) -> MentionTypes:
        """Verify that `mention_type` matches provided `mention_data`.

        Arguments:
            cls: this class.
            mention_type: mention type that should be consistent with data.
            values: verified data.

        Returns:
            Checked mention type.

        Raises:
            ValueError: raised if mention_type does not corresponds with data.
        """
        mention_data = values.get("mention_data")
        if mention_data is None:
            raise ValueError("no `mention_data`, perhaps this entity isn't a mention")

        user_mention_types = {MentionTypes.user, MentionTypes.contact}
        chat_mention_types = {MentionTypes.chat, MentionTypes.channel}

        if isinstance(mention_data, UserMention):
            if mention_type in user_mention_types:
                return mention_type
            raise ValueError(
                "mention_type for provided mention_data is wrong, accepted: {0}",
                user_mention_types,
            )

        if mention_type in chat_mention_types:
            return mention_type
        raise ValueError(
            "mention_type for provided mention_data is wrong, accepted: {0}",
            chat_mention_types,
        )



"""Typings for test client and mocks."""

from typing import Union

from botx.clients.methods.v2.bots import token
from botx.clients.methods.v3.chats import (
    add_admin_role,
    add_user,
    create,
    info,
    remove_user,
    stealth_disable,
    stealth_set,
)
from botx.clients.methods.v3.command import command_result
from botx.clients.methods.v3.events import edit_event
from botx.clients.methods.v3.notification import direct_notification, notification
from botx.clients.methods.v3.users import by_email, by_huid, by_login

APIMessage = Union[
    command_result.CommandResult,
    notification.Notification,
    direct_notification.NotificationDirect,
    edit_event.EditEvent,
]

APIRequest = Union[
    # V2
    # bots
    token.Token,
    # V3
    # chats
    add_admin_role.AddAdminRole,
    add_user.AddUser,
    info.Info,
    remove_user.RemoveUser,
    stealth_disable.StealthDisable,
    stealth_set.StealthSet,
    create.Create,
    # command
    command_result.CommandResult,
    # notification
    notification.Notification,
    direct_notification.NotificationDirect,
    # events
    edit_event.EditEvent,
    # users
    by_huid.ByHUID,
    by_email.ByEmail,
    by_login.ByLogin,
]



"""Definition for "chat creating disallowed" error."""
from typing import NoReturn
from uuid import UUID

from httpx import Response
from pydantic import BaseModel

from botx.clients.methods.base import APIErrorResponse, BotXMethod
from botx.exceptions import BotXAPIError


class ChatCreationDisallowedError(BotXAPIError):
    """Error for raising when bot is not allowed to create chats."""

    message_template = "bot {bot_id} is not allowed to create chats"

    #: ID of bot that sent request.
    bot_id: UUID


class ChatCreationDisallowedData(BaseModel):
    """Data for error when bot is not allowed to create chats."""

    #: ID of bot that sent request.
    bot_id: UUID


def handle_error(method: BotXMethod, response: Response) -> NoReturn:
    """Handle "chat creating disallowed" error response.

    Arguments:
        method: method which was made before error.
        response: HTTP response from BotX API.

    Raises:
        ChatCreationDisallowedError: raised always.
    """
    parsed_response = APIErrorResponse[ChatCreationDisallowedData].parse_obj(
        response.json(),
    )
    error_data = parsed_response.error_data
    raise ChatCreationDisallowedError(
        url=method.url,
        method=method.http_method,
        response_content=response.content,
        status_content=response.status_code,
        bot_id=error_data.bot_id,
    )



"""Entities for users."""

from typing import List, Optional
from uuid import UUID

from pydantic import BaseModel

from botx.models.enums import UserKinds


class UserInChatCreated(BaseModel):
    """User that can be included in data in `system:chat_created` event."""

    #: user HUID.
    huid: UUID

    #: type of user.
    user_kind: UserKinds

    #: user username.
    name: Optional[str]

    #: is user administrator in chat.
    admin: bool


class UserFromSearch(BaseModel):
    """User from search request."""

    #: HUID of user from search.
    user_huid: UUID

    #: AD login of user.
    ad_login: Optional[str]

    # AD domain of user.
    ad_domain: Optional[str]

    #: visible username.
    name: str

    #: user's company.
    company: Optional[str]

    #: user's position.
    company_position: Optional[str]

    #: user's department.
    department: Optional[str]

    #: user's emails.
    emails: List[str]


class UserFromChatSearch(BaseModel):
    """User from chat search request."""

    #: is user admin of chat.
    admin: bool

    #: HUID of user.
    user_huid: UUID

    #: type of user.
    user_kind: UserKinds



"""Dependant model and transforming functions."""
from __future__ import annotations

import inspect
from dataclasses import field
from typing import Any, Callable, Dict, List, Optional

from pydantic.dataclasses import dataclass
from pydantic.utils import lenient_issubclass

from botx.bots import bots
from botx.clients.clients import async_client, sync_client
from botx.dependencies import inspecting
from botx.models.messages.message import Message

WRONG_PARAM_TYPE_ERROR_TEXT = (
    "Param {0} of {1} can only be a dependency, message, bot or client, got: {2}"
)

CacheKey = Optional[Callable]
DependenciesCache = Dict[CacheKey, Any]


@dataclass
class Depends:
    """Stores dependency callable."""

    #: callable object that will be used in handlers or other dependencies instances.
    dependency: Callable[..., Any]

    #: use cache for dependency.
    use_cache: bool = True


@dataclass
class Dependant:
    """Main model that contains all necessary data for solving dependencies."""

    #: list of sub-dependencies for this dependency.
    dependencies: List[Dependant] = field(default_factory=list)

    #: name of dependency.
    name: Optional[str] = None

    #: callable object that will solve dependency.
    call: Optional[Callable] = None

    #: param name for passing incoming [message][botx.models.messages.Message]
    message_param_name: Optional[str] = None

    #: param name for passing [bot][botx.bots.Bot] that handles command.
    bot_param_name: Optional[str] = None

    #: param name for passing [client][botx.clients.clients.async_client.AsyncClient].
    async_client_param_name: Optional[str] = None

    #: param name for passing [client][botx.clients.clients.sync_client.Client].
    sync_client_param_name: Optional[str] = None

    #: use cache for optimize solving performance.
    use_cache: bool = True

    # Save the cache key at creation to optimize performance
    #: storage for cache.
    cache_key: CacheKey = field(init=False)

    def __post_init__(self) -> None:
        """Init special fields."""
        self.cache_key = self.call


Dependant.__pydantic_model__.update_forward_refs()  # type: ignore  # noqa: WPS609


def get_param_sub_dependant(*, dependency_param: inspect.Parameter) -> Dependant:
    """Parse instance of parameter to get it as dependency.

    Arguments:
        dependency_param: param for which sub dependency should be retrieved.

    Returns:
        Object that will be used in solving dependency.
    """
    depends: Depends = dependency_param.default
    dependency = depends.dependency

    return get_dependant(
        call=dependency, name=dependency_param.name, use_cache=depends.use_cache,
    )


def get_dependant(
    *, call: Callable, name: Optional[str] = None, use_cache: bool = True,
) -> Dependant:
    """Get dependant instance from passed callable object.

    Arguments:
        call: callable object that will be parsed to get required parameters and
            sub dependencies.
        name: name for dependency.
        use_cache: use cache for optimize solving performance.

    Returns:
        Object that will be used in solving dependency.

    Raises:
        ValueError: raised if param is not Dependant or special type.
    """
    dependant = Dependant(call=call, name=name, use_cache=use_cache)
    for dependency_param in inspecting.get_typed_signature(call).parameters.values():
        if isinstance(dependency_param.default, Depends):
            dependant.dependencies.append(
                get_param_sub_dependant(dependency_param=dependency_param),
            )
            continue

        is_special_param = add_special_param_to_dependency(
            dependency_param=dependency_param, dependant=dependant,
        )
        if is_special_param:
            continue

        raise ValueError(
            WRONG_PARAM_TYPE_ERROR_TEXT.format(
                dependency_param.name, call, dependency_param.annotation,
            ),
        )

    return dependant


def add_special_param_to_dependency(
    *, dependency_param: inspect.Parameter, dependant: Dependant,
) -> bool:
    """Check if param is non field object that should be passed into callable.

    Arguments:
        dependency_param: param that should be checked.
        dependant: dependency which field would be filled with required param name.

    Returns:
        Result of check.
    """
    if lenient_issubclass(dependency_param.annotation, bots.Bot):
        dependant.bot_param_name = dependency_param.name
        return True
    elif lenient_issubclass(dependency_param.annotation, Message):
        dependant.message_param_name = dependency_param.name
        return True
    elif lenient_issubclass(dependency_param.annotation, async_client.AsyncClient):
        dependant.async_client_param_name = dependency_param.name
        return True
    elif lenient_issubclass(dependency_param.annotation, sync_client.Client):
        dependant.sync_client_param_name = dependency_param.name
        return True

    return False



"""Mixin for shortcut for notification resource requests."""

from typing import Optional, Sequence
from uuid import UUID

from botx import converters
from botx.bots.mixins.requests.call_protocol import BotXMethodCallProtocol
from botx.clients.methods.v3.notification.direct_notification import NotificationDirect
from botx.clients.methods.v3.notification.notification import Notification
from botx.clients.types.message_payload import ResultPayload
from botx.clients.types.options import ResultOptions
from botx.models.messages.sending.credentials import SendingCredentials
from botx.models.messages.sending.payload import MessagePayload


class NotificationRequestsMixin:
    """Mixin for shortcut for notification resource requests."""

    async def send_notification(
        self: BotXMethodCallProtocol,
        credentials: SendingCredentials,
        payload: MessagePayload,
        group_chat_ids: Optional[Sequence[UUID]] = None,
    ) -> None:
        """Send notifications into chat.

        Arguments:
            credentials: credentials for making request.
            payload: payload for notification.
            group_chat_ids: IDS of chats into which message should be sent.
        """
        if group_chat_ids is not None:
            chat_ids = converters.optional_sequence_to_list(group_chat_ids)
        elif credentials.chat_id:
            chat_ids = [credentials.chat_id]
        else:
            chat_ids = []

        await self.call_method(
            Notification(
                group_chat_ids=chat_ids,
                result=ResultPayload(
                    body=payload.text,
                    metadata=payload.metadata,
                    bubble=payload.markup.bubbles,
                    keyboard=payload.markup.keyboard,
                    mentions=payload.options.mentions,
                ),
                recipients=payload.options.recipients,
                file=payload.file,
                opts=ResultOptions(
                    silent_response=payload.options.silent_response,
                    notification_opts=payload.options.notifications,
                ),
            ),
            credentials=credentials,
        )

    async def send_direct_notification(
        self: BotXMethodCallProtocol,
        credentials: SendingCredentials,
        payload: MessagePayload,
    ) -> UUID:
        """Send notification into chat.

        Arguments:
            credentials: credentials for making request.
            payload: payload for notification.

        Returns:
            ID sent message.

        Raises:
            ValueError: raised if chat_id wasn't provided
        """
        if not credentials.chat_id:
            raise ValueError("chat_id is required to send direct notification")

        return await self.call_method(
            NotificationDirect(
                group_chat_id=credentials.chat_id,
                event_sync_id=credentials.message_id,
                result=ResultPayload(
                    body=payload.text,
                    metadata=payload.metadata,
                    bubble=payload.markup.bubbles,
                    keyboard=payload.markup.keyboard,
                    mentions=payload.options.mentions,
                ),
                recipients=payload.options.recipients,
                file=payload.file,
                opts=ResultOptions(
                    silent_response=payload.options.silent_response,
                    notification_opts=payload.options.notifications,
                ),
            ),
            credentials=credentials,
        )



"""Payload for messages."""
from __future__ import annotations

from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field

from botx.models.constants import MAXIMUM_TEXT_LENGTH
from botx.models.files import File
from botx.models.mentions import Mention
from botx.models.messages.sending.markup import MessageMarkup
from botx.models.messages.sending.options import MessageOptions, NotificationOptions
from botx.models.typing import BubbleMarkup, KeyboardMarkup


class MessagePayload(BaseModel):
    """Message payload configuration."""

    #: message text.
    text: str = Field("", max_length=MAXIMUM_TEXT_LENGTH)

    #: message metadata.
    metadata: Dict[str, Any] = {}

    #: attached to message file.
    file: Optional[File] = None

    #: message markup.
    markup: MessageMarkup = MessageMarkup()

    #: message configuration.
    options: MessageOptions = MessageOptions()


class UpdatePayload(BaseModel):
    """Payload for message edition."""

    #: new message text.
    text: Optional[str] = Field(None, max_length=MAXIMUM_TEXT_LENGTH)

    #: new message bubbles.
    keyboard: Optional[KeyboardMarkup] = None

    #: new message keyboard.
    bubbles: Optional[BubbleMarkup] = None

    #: new message mentions.
    mentions: Optional[List[Mention]] = None

    #: new message options.
    opts: Optional[NotificationOptions] = None

    @property
    def markup(self) -> MessageMarkup:
        """Markup for edited message."""
        return MessageMarkup(bubbles=self.bubbles or [], keyboard=self.keyboard or [])

    def set_markup(self, markup: MessageMarkup) -> None:
        """Markup for edited message.

        Arguments:
            markup: markup that should be applied to payload.
        """
        self.bubbles = markup.bubbles
        self.keyboard = markup.keyboard

    @classmethod
    def from_sending_payload(cls, payload: MessagePayload) -> UpdatePayload:
        """Create new update payload from existing payload for new message.

        Arguments:
            payload: payload that can be used for sending new message.

        Returns:
            Created payload for update.
        """
        update = cls()
        update.text = payload.text or None
        update.set_markup(payload.markup)
        update.mentions = payload.options.mentions
        return update



"""Lifespan mixin for bot."""

import asyncio
from typing import List, Set

from botx.concurrency import callable_to_coroutine
from botx.typing import BotLifespanEvent

try:
    from typing import Protocol  # noqa: WPS433
except ImportError:
    from typing_extensions import (  # type: ignore  # noqa: WPS433, WPS440, F401
        Protocol,
    )


class LifespanMixin:
    """Lifespan events mixin for bot."""

    #: currently running tasks.
    tasks: Set[asyncio.Future]

    #: startup events.
    startup_events: List[BotLifespanEvent]

    #: shutdown events.
    shutdown_events: List[BotLifespanEvent]

    async def start(self) -> None:
        """Run all startup events and other initialization stuff."""
        for event in self.startup_events:
            await callable_to_coroutine(event, self)

    async def shutdown(self) -> None:
        """Wait for all running handlers shutdown."""
        await self.wait_current_handlers()

        for event in self.shutdown_events:
            await callable_to_coroutine(event, self)

    async def wait_current_handlers(self) -> None:
        """Wait until all current tasks are done."""
        if self.tasks:
            tasks, _ = await asyncio.wait(
                self.tasks, return_when=asyncio.ALL_COMPLETED,
            )
            for task in tasks:
                task.result()

        self.tasks.clear()



"""Definition for "chat not found" error."""
from typing import NoReturn
from uuid import UUID

from httpx import Response
from pydantic import BaseModel

from botx.clients.methods.base import APIErrorResponse, BotXMethod
from botx.exceptions import BotXAPIError


class ChatNotFoundError(BotXAPIError):
    """Error for raising when chat not found."""

    message_template = "chat {group_chat_id} not found"

    #: ID of chat that was requested.
    group_chat_id: UUID


class ChatNotFoundData(BaseModel):
    """Data for error when chat not found."""

    #: ID of chat that was requested.
    group_chat_id: UUID


def handle_error(method: BotXMethod, response: Response) -> NoReturn:
    """Handle "chat creation error" error response.

    Arguments:
        method: method which was made before error.
        response: HTTP response from BotX API.

    Raises:
        ChatNotFoundError: raised always.
    """
    parsed_response = APIErrorResponse[ChatNotFoundData].parse_obj(response.json())

    error_data = parsed_response.error_data
    raise ChatNotFoundError(
        url=method.url,
        method=method.http_method,
        response_content=response.content,
        status_content=response.status_code,
        group_chat_id=error_data.group_chat_id,
    )



"""Definition for mixin that defines BotX API methods."""
from typing import Optional, TypeVar, cast
from uuid import UUID

from botx.bots.mixins.requests import bots, chats, command, events, notification, users
from botx.clients.clients.async_client import AsyncClient
from botx.clients.methods.base import BotXMethod
from botx.clients.methods.v2.bots.token import Token
from botx.models.credentials import ExpressServer, ServerCredentials
from botx.models.messages.sending.credentials import SendingCredentials

ResponseT = TypeVar("ResponseT")

try:
    from typing import Protocol  # noqa: WPS433
except ImportError:
    from typing_extensions import (  # type: ignore  # noqa: WPS433, WPS440, F401
        Protocol,
    )


class CredentialsSearchProtocol(Protocol):
    """Protocol for search token in local credentials."""

    def get_token_for_cts(self, host: str) -> str:
        """Search token in local credentials."""

    def get_cts_by_host(self, host: str) -> ExpressServer:
        """Get CTS by host."""


# A lot of base classes since it's mixin for all shorthands for BotX API requests
class BotXRequestsMixin(  # noqa: WPS215
    bots.BotsRequestsMixin,
    chats.ChatsRequestsMixin,
    command.CommandRequestsMixin,
    events.EventsRequestsMixin,
    notification.NotificationRequestsMixin,
    users.UsersRequestsMixin,
):
    """Mixin that defines methods for communicating with BotX API."""

    client: AsyncClient

    async def call_method(  # noqa: WPS211
        self,
        method: BotXMethod[ResponseT],
        *,
        host: Optional[str] = None,
        token: Optional[str] = None,
        bot_id: Optional[UUID] = None,
        credentials: Optional[SendingCredentials] = None,
    ) -> ResponseT:
        """Call method with async client.

        Arguments:
            method: method that should be user for request.
            host: host where request should be sent.
            token: token for method.
            bot_id: ID of bot that send request.
            credentials: credentials for making request.

        Returns:
            Response for method.
        """
        if not isinstance(method, Token):
            host = cast(str, credentials.host if credentials else host)
            await _fill_token(
                self.client,
                host,
                cast(UUID, credentials.bot_id if credentials else bot_id),
                cast(CredentialsSearchProtocol, self).get_cts_by_host(host),
            )

        if credentials is not None:
            host = cast(str, credentials.host)
            method.configure(
                host=host,
                token=cast(CredentialsSearchProtocol, self).get_token_for_cts(host),
            )
        else:
            method.configure(host=host or method.host, token=token or method.token)

        return await self.client.call(method)


async def _fill_token(
    client: AsyncClient, host: str, bot_id: UUID, server: ExpressServer,
) -> None:
    if server.server_credentials is not None:
        return

    method = Token(bot_id=bot_id, signature=server.calculate_signature(bot_id))
    server.server_credentials = ServerCredentials(
        bot_id=bot_id, token=await client.call(method, host),
    )



"""Responses from BotX API."""
from uuid import UUID

from pydantic import BaseModel


class PushResult(BaseModel):
    """Entity that contains result from notification or command result push."""

    #: event id of pushed message.
    sync_id: UUID


class ChatCreatedResult(BaseModel):
    """Entity that contains result from chat creation."""

    #: id of created chat.
    chat_id: UUID



"""Mixin that defines handler decorator."""

from typing import Any, Callable, Optional, Sequence

from botx.collecting.collectors.collector import Collector
from botx.dependencies.models import Depends
from botx.models.enums import SystemEvents


class SystemEventsHandlerMixin:
    """Mixin that defines handler decorator."""

    collector: Collector

    def system_event(  # noqa: WPS211
        self,
        handler: Optional[Callable] = None,
        *,
        event: Optional[SystemEvents] = None,
        events: Optional[Sequence[SystemEvents]] = None,
        name: Optional[str] = None,
        dependencies: Optional[Sequence[Depends]] = None,
        dependency_overrides_provider: Any = None,
    ) -> Callable:
        """Register handler for system event.

        Arguments:
            handler: callable that will be used for executing handler.
            event: event for triggering this handler.
            events: a sequence of events that will trigger handler.
            name: optional name for handler that will be used in generating body.
            dependencies: sequence of dependencies that should be executed before
                handler.
            dependency_overrides_provider: mock of callable for handler.

        Returns:
            Passed in `handler` callable.
        """
        return self.collector.system_event(
            handler=handler,
            event=event,
            events=events,
            name=name,
            dependencies=dependencies,
            dependency_overrides_provider=dependency_overrides_provider,
        )

    def chat_created(
        self,
        handler: Optional[Callable] = None,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
        dependency_overrides_provider: Any = None,
    ) -> Callable:
        """Register handler for `system:chat_created` event.

        Arguments:
            handler: callable that will be used for executing handler.
            dependencies: sequence of dependencies that should be executed before
                handler.
            dependency_overrides_provider: mock of callable for handler.

        Returns:
            Passed in `handler` callable.
        """
        return self.collector.chat_created(
            handler=handler,
            dependencies=dependencies,
            dependency_overrides_provider=dependency_overrides_provider,
        )

    def file_transfer(
        self,
        handler: Optional[Callable] = None,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
        dependency_overrides_provider: Any = None,
    ) -> Callable:
        """Register handler for `file_transfer` event.

        Arguments:
            handler: callable that will be used for executing handler.
            dependencies: sequence of dependencies that should be executed before
                handler.
            dependency_overrides_provider: mock of callable for handler.

        Returns:
            Passed in `handler` callable.
        """
        return self.collector.file_transfer(
            handler=handler,
            dependencies=dependencies,
            dependency_overrides_provider=dependency_overrides_provider,
        )

    def added_to_chat(
        self,
        handler: Optional[Callable] = None,
        *,
        dependencies: Optional[Sequence[Depends]] = None,
        dependency_overrides_provider: Any = None,
    ) -> Callable:
        """Register handler for `added_to_chat` event.

        Arguments:
            handler: callable that will be used for executing handler.
            dependencies: sequence of dependencies that should be executed before
                handler.
            dependency_overrides_provider: mock of callable for handler.

        Returns:
            Passed in `handler` callable.
        """
        return self.collector.system_event(
            handler=handler,
            event=SystemEvents.added_to_chat,
            name=SystemEvents.added_to_chat.value,
            dependencies=dependencies,
            dependency_overrides_provider=dependency_overrides_provider,
        )



"""Endpoints for bots resource."""

from starlette.requests import Request
from starlette.responses import Response

from botx.clients.methods.base import APIResponse
from botx.clients.methods.v2.bots.token import Token
from botx.testing.botx_mock.asgi.messages import add_request_to_collection
from botx.testing.botx_mock.asgi.responses import PydanticResponse
from botx.testing.botx_mock.binders import bind_implementation_to_method


@bind_implementation_to_method(Token)
async def get_token(request: Request) -> Response:
    """Handle retrieving token from BotX API request.

    Arguments:
        request: starlette request for route.

    Returns:
        Return response with new token.
    """
    request_data = {**request.path_params, **request.query_params}
    payload = Token(**request_data)
    add_request_to_collection(request, payload)
    return PydanticResponse(APIResponse[str](result="real token"))



"""A little python framework for building bots for Express."""

from loguru import logger

from botx.bots.bots import Bot
from botx.clients.clients.async_client import AsyncClient
from botx.clients.clients.sync_client import Client
from botx.collecting.collectors.collector import Collector
from botx.dependencies.injection_params import Depends
from botx.exceptions import BotXAPIError, DependencyFailure, ServerUnknownError
from botx.models.buttons import BubbleElement, KeyboardElement
from botx.models.credentials import ExpressServer, ServerCredentials
from botx.models.enums import (
    ChatTypes,
    CommandTypes,
    EntityTypes,
    MentionTypes,
    Statuses,
    SystemEvents,
    UserKinds,
)
from botx.models.errors import BotDisabledErrorData, BotDisabledResponse
from botx.models.events import ChatCreatedEvent
from botx.models.files import File
from botx.models.forwards import Forward
from botx.models.mentions import ChatMention, Mention, UserMention
from botx.models.menu import Status
from botx.models.messages.incoming_message import Entity, IncomingMessage
from botx.models.messages.message import Message
from botx.models.messages.sending.credentials import SendingCredentials
from botx.models.messages.sending.markup import MessageMarkup
from botx.models.messages.sending.message import SendingMessage
from botx.models.messages.sending.options import MessageOptions, NotificationOptions
from botx.models.messages.sending.payload import MessagePayload, UpdatePayload
from botx.testing.building.builder import MessageBuilder

try:
    from botx.testing.testing_client.client import TestClient  # noqa: WPS433
except ImportError:
    TestClient = None  # type: ignore  # noqa: WPS440

__all__ = (
    # bots
    "Bot",
    # collecting
    "Collector",
    # clients
    "AsyncClient",
    "Client",
    # exceptions
    "BotXAPIError",
    "ServerUnknownError",
    "DependencyFailure",
    # DI
    "Depends",
    # models
    # murkup
    "BubbleElement",
    "KeyboardElement",
    # credentials
    "ExpressServer",
    "ServerCredentials",
    # enums
    "Statuses",
    "UserKinds",
    "ChatTypes",
    "CommandTypes",
    "SystemEvents",
    "EntityTypes",
    "MentionTypes",
    # errors
    "BotDisabledErrorData",
    "BotDisabledResponse",
    # events
    "ChatCreatedEvent",
    # files
    "File",
    # mentions
    "Mention",
    "ChatMention",
    "UserMention",
    # status
    "Status",
    # messages
    # handler message
    "Message",
    # incoming
    "IncomingMessage",
    "Entity",
    "Forward",
    # sending
    "SendingCredentials",
    "SendingMessage",
    "MessageMarkup",
    "MessageOptions",
    "NotificationOptions",
    "MessagePayload",
    "UpdatePayload",
    # testing
    "TestClient",
    "MessageBuilder",
)

logger.disable("botx")



"""Shape that is used for messages from bot."""
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field

from botx.models.constants import MAXIMUM_TEXT_LENGTH
from botx.models.enums import Statuses
from botx.models.mentions import Mention
from botx.models.typing import BubbleMarkup, KeyboardMarkup

try:
    from typing import Literal  # noqa: WPS433
except ImportError:
    from typing_extensions import (  # type: ignore  # noqa: WPS433, WPS440, F401
        Literal,
    )


class ResultPayload(BaseModel):
    """Data that is sent when bot answers on command or send notification."""

    #: status of operation.
    status: Literal[Statuses.ok] = Statuses.ok

    #: body for new message from bot.
    body: str = Field("", max_length=MAXIMUM_TEXT_LENGTH)

    #: message metadata.
    metadata: Dict[str, Any] = {}

    #: keyboard that will be used for new message.
    keyboard: KeyboardMarkup = []

    #: bubble elements that will be showed under new message.
    bubble: BubbleMarkup = []

    #: mentions that BotX API will append before new message text.
    mentions: List[Mention] = []


class UpdatePayload(BaseModel):
    """Data that is sent when bot updates message."""

    #: status of operation.
    status: Literal[Statuses.ok] = Statuses.ok

    #: new body in message.
    body: Optional[str] = Field(None, max_length=MAXIMUM_TEXT_LENGTH)

    #: new keyboard that will be used for new message.
    keyboard: Optional[KeyboardMarkup] = None

    #: new bubble elements that will be showed under new message.
    bubble: Optional[BubbleMarkup] = None

    #: new mentions that BotX API will append before new message text.
    mentions: Optional[List[Mention]] = None



"""Entities for chats."""

from typing import List, Optional
from uuid import UUID

from pydantic import BaseModel

from botx.models.enums import ChatTypes
from botx.models.users import UserFromChatSearch


class ChatFromSearch(BaseModel):
    """Chat from search request."""

    #: name of chat.
    name: str

    #: description of chat
    description: Optional[str]

    #: type of chat.
    chat_type: ChatTypes

    #: HUID of chat creator.
    creator: UUID

    #: ID of chat.
    group_chat_id: UUID

    #: users in chat.
    members: List[UserFromChatSearch]



"""Definition for mixin with handler decorator."""
from functools import partial
from typing import Any, Callable, List, Optional, Sequence, Union, cast

from botx import converters
from botx.dependencies.models import Depends

try:
    from typing import Protocol  # noqa: WPS433
except ImportError:
    from typing_extensions import (  # type: ignore  # noqa: WPS433, WPS440, F401
        Protocol,
    )


class AddHandlerProtocol(Protocol):
    """Protocol for definition add_handler method."""

    def add_handler(  # noqa: WPS211
        self,
        handler: Callable,
        *,
        body: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        full_description: Optional[str] = None,
        include_in_status: Union[bool, Callable] = True,
        dependencies: Optional[Sequence[Depends]] = None,
        dependency_overrides_provider: Any = None,
    ) -> None:
        """Create new handler from passed arguments and store it inside."""


class HandlerDecoratorProtocol(Protocol):
    """Protocol for definition handler decorator."""

    def handler(  # noqa: WPS211
        self,
        handler: Optional[Callable] = None,
        *,
        command: Optional[str] = None,
        commands: Optional[Sequence[str]] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        full_description: Optional[str] = None,
        include_in_status: Union[bool, Callable] = True,
        dependencies: Optional[Sequence[Depends]] = None,
        dependency_overrides_provider: Any = None,
    ) -> Callable:
        """Add new handler to collector."""


class HandlerMixin:
    """Mixin that defines handler decorator."""

    def handler(  # noqa: WPS211
        self: AddHandlerProtocol,
        handler: Optional[Callable] = None,
        *,
        command: Optional[str] = None,
        commands: Optional[Sequence[str]] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        full_description: Optional[str] = None,
        include_in_status: Union[bool, Callable] = True,
        dependencies: Optional[Sequence[Depends]] = None,
        dependency_overrides_provider: Any = None,
    ) -> Callable:
        """Add new handler to collector.

        !!! info
            If `include_in_status` is a function, then `body` argument will be checked
            for matching public commands style, like `/command`.

        Arguments:
            handler: callable that will be used for executing handler.
            command: body template that will trigger this handler.
            commands: list of body templates that will trigger this handler.
            name: optional name for handler that will be used in generating body.
            description: description for command that will be shown in bot's menu.
            full_description: full description that can be used for example in `/help`
                command.
            include_in_status: should this handler be shown in bot's menu, can be
                callable function with no arguments *(for now)*.
            dependencies: sequence of dependencies that should be executed before
                handler.
            dependency_overrides_provider: mock of callable for handler.

        Returns:
            Passed in `handler` callable.
        """
        if handler:
            handler_commands: List[
                Optional[str]
            ] = converters.optional_sequence_to_list(commands)

            if command and commands:
                handler_commands.insert(0, command)
            elif not commands:
                handler_commands = [command]

            for command_body in handler_commands:
                self.add_handler(
                    body=command_body,
                    handler=handler,
                    name=name,
                    description=description,
                    full_description=full_description,
                    include_in_status=include_in_status,
                    dependencies=dependencies,
                    dependency_overrides_provider=dependency_overrides_provider,
                )

            return handler

        return partial(
            cast(HandlerDecoratorProtocol, self).handler,
            command=command,
            commands=commands,
            name=name,
            description=description,
            full_description=full_description,
            include_in_status=include_in_status,
            dependencies=dependencies,
            dependency_overrides_provider=dependency_overrides_provider,
        )



"""Definition for collector."""
from typing import Any, Optional, Sequence

from loguru import logger
from pydantic.dataclasses import dataclass

from botx.collecting.collectors.base import BaseCollector
from botx.collecting.collectors.mixins.default import DefaultHandlerMixin
from botx.collecting.collectors.mixins.handler import HandlerMixin
from botx.collecting.collectors.mixins.hidden import HiddenHandlerMixin
from botx.collecting.collectors.mixins.system_events import SystemEventsHandlerMixin
from botx.dependencies.models import Depends
from botx.exceptions import NoMatchFound
from botx.models.messages.message import Message


@dataclass
class Collector(  # noqa: WPS215
    HandlerMixin,
    DefaultHandlerMixin,
    HiddenHandlerMixin,
    SystemEventsHandlerMixin,
    BaseCollector,
):
    """Collector for different handlers."""

    async def __call__(self, message: Message) -> None:
        """Find handler and execute it.

        Arguments:
            message: incoming message that will be passed to handler.
        """
        await self.handle_message(message)

    def include_collector(
        self,
        collector: "Collector",
        *,
        dependencies: Optional[Sequence[Depends]] = None,
    ) -> None:
        """Include handlers from another collector into this one.

        Arguments:
            collector: collector from which handlers should be copied.
            dependencies: optional sequence of dependencies for handlers for this
                collector.

        Raises:
            AssertionError: raised if both collectors defines default handlers.
        """
        if self.default_message_handler and collector.default_message_handler:
            raise AssertionError("only one default handler can be applied")

        if collector.default_message_handler:
            self._add_default_handler(collector.default_message_handler, dependencies)

        self._add_handlers(collector.handlers, dependencies)

    def command_for(self, *args: Any) -> str:
        """Find handler and build a command string using passed body query_params.

        Arguments:
            args: sequence of elements where first element should be name of handler.

        Returns:
            Command string.

        Raises:
            TypeError: raised no arguments passed.
        """
        if not len(args):
            raise TypeError("missing handler name as the first argument")

        return self.handler_for(args[0]).command_for(*args)

    async def handle_message(self, message: Message) -> None:
        """Find handler and execute it.

        Arguments:
            message: incoming message that will be passed to handler.

        Raises:
            NoMatchFound: raised if no handler for message found.
        """
        for handler in self.sorted_handlers:
            if handler.matches(message):
                logger.bind(botx_collector=True).info(
                    "botx => {0}: {1}".format(handler.name, message.command.command),
                )
                await handler(message)
                return

        if self.default_message_handler:
            await self.default_message_handler(message)
        else:
            raise NoMatchFound(search_param=message.body)



"""Method for searching user by his email."""
import httpx

from botx.clients.methods.base import AuthorizedBotXMethod
from botx.clients.methods.errors import user_not_found
from botx.models.users import UserFromSearch


class ByEmail(AuthorizedBotXMethod[UserFromSearch]):
    """Method for searching user by his email."""

    __url__ = "/api/v3/botx/users/by_email"
    __method__ = "GET"
    __returning__ = UserFromSearch
    __errors_handlers__ = {httpx.codes.NOT_FOUND: user_not_found.handle_error}

    #: email to search
    email: str



"""Shared config for pydantic dataclasses."""

from pydantic import BaseConfig


class BotXDataclassConfig(BaseConfig):
    """Config for pydantic dataclasses that allows custom types."""

    arbitrary_types_allowed = True



"""Method for disabling stealth in chat."""
from uuid import UUID

import httpx

from botx.clients.methods.base import AuthorizedBotXMethod
from botx.clients.methods.errors import bot_is_not_admin, chat_not_found


class StealthDisable(AuthorizedBotXMethod[bool]):
    """Method for disabling stealth in chat."""

    __url__ = "/api/v3/botx/chats/stealth_disable"
    __method__ = "POST"
    __returning__ = bool
    __errors_handlers__ = {
        httpx.codes.FORBIDDEN: bot_is_not_admin.handle_error,
        httpx.codes.NOT_FOUND: chat_not_found.handle_error,
    }

    #: ID of chat where stealth should be disabled.
    group_chat_id: UUID



"""Helpers for execution functions as coroutines."""

import asyncio
import contextvars
import functools
import inspect
from typing import Any, Callable, Coroutine


def is_awaitable_object(call: Callable) -> bool:
    """Check if object is an awaitable or an object which __call__ method is awaitable.

    Arguments:
        call: callable for checking.

    Returns:
        Result of check.
    """
    if is_awaitable(call):
        return True
    call = getattr(call, "__call__", None)  # noqa: B004
    return asyncio.iscoroutinefunction(call)


def is_awaitable(call: Callable) -> bool:
    """Check if function returns awaitable object.

    Arguments:
        call: function that should be checked.

    Returns:
        Result of check.
    """
    if inspect.isfunction(call) or inspect.ismethod(call):
        return asyncio.iscoroutinefunction(call)
    return False


async def run_in_threadpool(call: Callable, *args: Any, **kwargs: Any) -> Any:
    """Run regular function (not a coroutine) as awaitable coroutine.

    Arguments:
        call: function that should be called as coroutine.
        args: positional arguments for the function.
        kwargs: keyword arguments for the function.

    Returns:
        Result of function call.
    """
    loop = asyncio.get_event_loop()
    child = functools.partial(call, *args, **kwargs)
    context = contextvars.copy_context()
    call = context.run
    args = (child,)
    return await loop.run_in_executor(None, call, *args)


def callable_to_coroutine(func: Callable, *args: Any, **kwargs: Any) -> Coroutine:
    """Transform callable to coroutine.

    Arguments:
        func: function that can be sync or async and should be transformed into
            corouine.
        args: positional arguments for this function.
        kwargs: key arguments for this function.

    Returns:
        Coroutine object from passed callable.
    """
    if is_awaitable_object(func):
        return func(*args, **kwargs)

    return run_in_threadpool(func, *args, **kwargs)


def run_in_blocking_loop(call: Callable, *args: Any, **kwargs: Any) -> Any:
    """Run coroutine with loop blocking.

    Arguments:
        call: function that should be called in loop until complete.
        args: function arguments.
        kwargs: function key arguments.

    Returns:
        Result of function call.
    """
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()

    return loop.run_until_complete(callable_to_coroutine(call, *args, **kwargs))


def async_to_sync(func: Callable) -> Callable:
    """Convert asynchronous function to blocking.

    Arguments:
        func: function that should be converted.

    Returns:
        Converted function.
    """
    return functools.partial(run_in_blocking_loop, func)



"""Definition for base collector."""

from dataclasses import InitVar, field
from typing import Any, Callable, List, Optional, Sequence, Union

from pydantic.dataclasses import dataclass

from botx import converters
from botx.collecting.handlers.handler import Handler
from botx.collecting.handlers.name_generators import (
    get_body_from_name,
    get_name_from_callable,
)
from botx.dependencies import models as deps
from botx.dependencies.models import Depends
from botx.exceptions import NoMatchFound


def _get_sorted_handlers(handlers: List[Handler]) -> List[Handler]:
    return sorted(handlers, key=lambda handler: len(handler.body), reverse=True)


def _combine_dependencies(
    *dependencies: Optional[Sequence[deps.Depends]],
) -> List[deps.Depends]:
    result_dependencies = []
    for deps_sequence in dependencies:
        result_dependencies.extend(converters.optional_sequence_to_list(deps_sequence))

    return result_dependencies


def _check_new_handler_restrictions(
    body: str, name: Optional[str], handler: Callable, existed_handler: Handler,
) -> None:
    handler_executor = existed_handler.handler
    handler_name = existed_handler.name

    if body == existed_handler.body:
        raise AssertionError("handler with body {0} already registered".format(body))

    handler_registered = handler == handler_executor and name == handler_name
    if name == existed_handler.name and not handler_registered:
        raise AssertionError("handler with name {0} already registered".format(name))


@dataclass
class BaseCollector:
    """Base collector."""

    default: InitVar[Handler] = None

    #: registered handlers on this collector handlers in order of adding.
    handlers: List[Handler] = field(default_factory=list)

    #: handler that will be used for handling non matched message.
    default_message_handler: Optional[Handler] = None

    #: background dependencies that will be executed for handlers.
    dependencies: Optional[Sequence[Depends]] = None

    #: overrider for dependencies.
    dependency_overrides_provider: Optional[Any] = None

    @property
    def sorted_handlers(self) -> List[Handler]:
        """Get added handlers sorted by bodies length.

        Returns:
            Sorted handlers.
        """
        return _get_sorted_handlers(self.handlers)

    def __post_init__(self, default: Optional[Handler]) -> None:
        """Initialize or update special fields.

        Arguments:
            default: callable that should be used as default handler.
        """
        handlers = self.handlers
        self.handlers = []
        self._add_handlers(converters.optional_sequence_to_list(handlers))

        if default is not None:
            self._add_default_handler(default)

    def handler_for(self, name: str) -> Handler:
        """Find handler in handlers of this bot.

        Arguments:
            name: name of handler that should be found.

        Returns:
            Handler that was found by name.

        Raises:
            NoMatchFound: raise if handler was not found.
        """
        for handler in self.handlers:
            if handler.name == name:
                return handler

        raise NoMatchFound(search_param=name)

    def add_handler(  # noqa: WPS211
        self,
        handler: Callable,
        *,
        body: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        full_description: Optional[str] = None,
        include_in_status: Union[bool, Callable] = True,
        dependencies: Optional[Sequence[deps.Depends]] = None,
        dependency_overrides_provider: Any = None,
    ) -> None:
        """Create new handler from passed arguments and store it inside.

        !!! info
            If `include_in_status` is a function, then `body` argument will be checked
            for matching public commands style, like `/command`.

        Arguments:
            handler: callable that will be used for executing handler.
            body: body template that will trigger this handler.
            name: optional name for handler that will be used in generating body.
            description: description for command that will be shown in bot's menu.
            full_description: full description that can be used for example in `/help`
                command.
            include_in_status: should this handler be shown in bot's menu, can be
                callable function with no arguments *(for now)*.
            dependencies: sequence of dependencies that should be executed before
                handler.
            dependency_overrides_provider: mock of callable for handler.
        """
        if body is None:
            name = name or get_name_from_callable(handler)
            body = get_body_from_name(name)

        for registered_handler in self.handlers:
            _check_new_handler_restrictions(body, name, handler, registered_handler)

        dep_override = (
            dependency_overrides_provider
            if dependency_overrides_provider is not None
            else self.dependency_overrides_provider
        )
        command_handler = Handler(
            body=body,
            handler=handler,
            name=name,  # type: ignore
            description=description,
            full_description=full_description,  # type: ignore
            include_in_status=include_in_status,
            dependencies=_combine_dependencies(self.dependencies, dependencies),
            dependency_overrides_provider=dep_override,
        )
        self.handlers.append(command_handler)

    def _add_handlers(
        self, handlers: List[Handler], dependencies: Optional[Sequence[Depends]] = None,
    ) -> None:
        for handler in handlers:
            self.add_handler(
                body=handler.body,
                handler=handler.handler,
                name=handler.name,
                description=handler.description,
                full_description=handler.full_description,
                include_in_status=handler.include_in_status,
                dependencies=handler.dependencies,
            )
            created_handler = self.handler_for(handler.name)
            created_handler.dependencies = _combine_dependencies(
                dependencies, created_handler.dependencies,
            )

    def _add_default_handler(
        self, default: Handler, dependencies: Optional[Sequence[Depends]] = None,
    ) -> None:
        default_dependencies = _combine_dependencies(
            self.dependencies, dependencies, default.dependencies,
        )
        self.default_message_handler = Handler(
            body=default.body,
            handler=default.handler,
            name=default.name,
            description=default.description,
            full_description=default.full_description,
            include_in_status=default.include_in_status,
            dependencies=default_dependencies,
            dependency_overrides_provider=self.dependency_overrides_provider,
        )



"""Definition for "messaging" error."""
from typing import NoReturn

from httpx import Response

from botx.clients.methods.base import BotXMethod
from botx.exceptions import BotXAPIError


class MessagingError(BotXAPIError):
    """Error for raising when there is messaging error."""

    message_template = "error from messaging service"


def handle_error(method: BotXMethod, response: Response) -> NoReturn:
    """Handle messaging error response.

    Arguments:
        method: method which was made before error.
        response: HTTP response from BotX API.

    Raises:
        MessagingError: raised always.
    """
    raise MessagingError(
        url=method.url,
        method=method.http_method,
        response=response.json(),
        status=response.status_code,
    )



"""Mixin for shortcut for events resource requests."""

from botx.bots.mixins.requests.call_protocol import BotXMethodCallProtocol
from botx.clients.methods.v3.events.edit_event import EditEvent
from botx.clients.types.message_payload import UpdatePayload
from botx.models.messages.sending.credentials import SendingCredentials
from botx.models.messages.sending.payload import UpdatePayload as SendingUpdatePayload


class EventsRequestsMixin:
    """Mixin that defines methods for communicating with BotX API."""

    async def update_message(
        self: BotXMethodCallProtocol,
        credentials: SendingCredentials,
        update: SendingUpdatePayload,
    ) -> None:
        """Change message by it's event id.

        Arguments:
            credentials: credentials that are used for sending message. *sync_id* is
                required for credentials.
            update: update of message content.

        Raises:
            ValueError: raised if sync_id wasn't provided
        """
        if not credentials.sync_id:
            raise ValueError("sync_id is required for message update")

        await self.call_method(
            EditEvent(
                sync_id=credentials.sync_id,
                result=UpdatePayload(
                    body=update.text,
                    keyboard=update.keyboard,
                    bubble=update.bubbles,
                    mentions=update.mentions,
                ),
            ),
            credentials=credentials,
        )



"""Pydantic models for bubbles and keyboard buttons."""

from typing import Optional

from pydantic import BaseModel, validator


class ButtonOptions(BaseModel):
    """Extra options for buttons, like disabling output by tap."""

    #: if True then text won't shown for user in messenger.
    silent: bool = True

    #: text to be shown in toast (show command body if `alert_text` is `None`).
    alert_text: Optional[str] = None

    #: show toast with `alert_text` when user press the button
    show_alert: bool = False


class Button(BaseModel):
    """Base class for ui element like bubble or keyboard button."""

    #: command that will be triggered by click on the element.
    command: str

    #: text that will be shown on the element.
    label: Optional[str] = None

    #: extra payload that will be stored in button and then received in new message.
    data: dict = {}  # noqa: WPS110

    #: options for button.
    opts: ButtonOptions = ButtonOptions()

    @validator("label", always=True)
    def label_as_command_if_none(
        cls, label: Optional[str], values: dict,  # noqa: N805, WPS110
    ) -> str:
        """Return command as label if it is `None`.

        Arguments:
            cls: passed button class.
            label: value that should be checked.
            values: all other validated_values checked before.

        Returns:
            Label for button.
        """
        return label or values["command"]

    @validator("data", always=True)
    def add_ui_flag_to_data(cls, button_data: dict) -> dict:  # noqa: N805
        """Return command data with set UI flag.

        Arguments:
            cls: passed button class.
            button_data: data passed to bot.

        Returns:
            Passed data with set "ui" flag to True if it wasn't set already.
        """
        button_data.setdefault("ui", True)
        return button_data


class BubbleElement(Button):
    """Bubble buttons that is shown under messages."""


class KeyboardElement(Button):
    """Keyboard buttons that are placed instead of real keyboard."""



"""Definition for "bot is not admin" error."""
from typing import NoReturn
from uuid import UUID

from httpx import Response
from pydantic import BaseModel

from botx.clients.methods.base import APIErrorResponse, BotXMethod
from botx.exceptions import BotXAPIError


class BotIsNotAdminError(BotXAPIError):
    """Error for raising when bot is not admin."""

    message_template = "bot {bot_id} is not admin of chat {group_chat_id}"

    #: ID of bot that sent request.
    bot_id: UUID

    #: ID of chat into which request was sent.
    group_chat_id: UUID


class BotIsNotAdminData(BaseModel):
    """Data for error when bot is not admin."""

    #: ID of sender (bot)
    sender: UUID

    #: ID of chat into which request was sent.
    group_chat_id: UUID


def handle_error(method: BotXMethod, response: Response) -> NoReturn:
    """Handle "bot is not admin" error response.

    Arguments:
        method: method which was made before error.
        response: HTTP response from BotX API.

    Raises:
        BotIsNotAdminError: raised always.
    """
    error_data = (
        APIErrorResponse[BotIsNotAdminData].parse_obj(response.json()).error_data
    )
    raise BotIsNotAdminError(
        url=method.url,
        method=method.http_method,
        response_content=response.content,
        status_content=response.status_code,
        bot_id=error_data.sender,
        group_chat_id=error_data.group_chat_id,
    )



"""Definition of base middleware class and some default middlewares."""

from typing import Callable, Dict, Optional, Type

from loguru import logger

from botx import concurrency
from botx.middlewares.base import BaseMiddleware
from botx.models import files
from botx.models.messages.message import Message
from botx.typing import AsyncExecutor, Executor


class ExceptionMiddleware(BaseMiddleware):
    """Custom middleware that is default and used to handle registered errors."""

    def __init__(self, executor: Executor) -> None:
        """Init middleware with required query_params.

        Arguments:
            executor: callable object that accept message and will be executed after
                middleware.
        """
        super().__init__(executor)
        self._exception_handlers: Dict[Type[Exception], Callable] = {}

    async def dispatch(self, message: Message, call_next: AsyncExecutor) -> None:
        """Wrap executor for catching exception or log them.

        Arguments:
            message: incoming message that will be passed to executor.
            call_next: next executor that should be called after this.
        """
        try:
            await call_next(message)
        except Exception as exc:
            await self._handle_error_in_handler(exc, message)

    def add_exception_handler(
        self, exc_class: Type[Exception], handler: Callable,
    ) -> None:
        """Register handler for specific exception in middleware.

        Arguments:
            exc_class: exception class that should be handled by middleware.
            handler: handler for exception.
        """
        self._exception_handlers[exc_class] = handler

    def _lookup_handler_for_exception(self, exc: Exception) -> Optional[Callable]:
        """Find handler for exception.

        Arguments:
            exc: catched exception for which handler should be found.

        Returns:
            Found handler or None.
        """
        for exc_cls in type(exc).mro():
            handler = self._exception_handlers.get(exc_cls)
            if handler:
                return handler

        return None

    async def _handle_error_in_handler(self, exc: Exception, message: Message) -> None:
        """Pass error back to handler if there is one or log error.

        Arguments:
            exc: exception that occurred.
            message: message on which exception occurred.
        """
        handler = self._lookup_handler_for_exception(exc)

        if handler is None:
            logger.bind(
                botx_error=True,
                payload=message.incoming_message.copy(
                    update={
                        "body": _convert_text_to_logs_format(message.body),
                        "file": _convert_file_to_logs_format(message.file),
                    },
                ).dict(),
            ).exception("uncaught {0} exception {1}", type(exc).__name__, exc)
            return

        await concurrency.callable_to_coroutine(handler, exc, message)


def _convert_text_to_logs_format(text: str) -> str:
    """Convert text into format that is suitable for logs.

    Arguments:
        text: text that should be formatted.

    Returns:
        Shape for logging in loguru.
    """
    max_log_text_length = 50
    start_text_index = 15
    end_text_index = 5

    return (
        "...".join((text[:start_text_index], text[-end_text_index:]))
        if len(text) > max_log_text_length
        else text
    )


def _convert_file_to_logs_format(file: Optional[files.File]) -> Optional[dict]:
    """Convert file to a new file that will be showed in logs.

    Arguments:
        file: file that should be converted.

    Returns:
        New file or nothing.
    """
    return file.copy(update={"data": "[file content]"}).dict() if file else None



"""Define Mixin that combines API and sending mixins and will be used in Bot."""
from typing import List

from botx.bots.mixins.requests.mixin import BotXRequestsMixin
from botx.bots.mixins.sending import SendingMixin
from botx.exceptions import ServerUnknownError
from botx.models.credentials import ExpressServer


class ClientsMixin(SendingMixin, BotXRequestsMixin):
    """Mixin that defines methods that are used for communicating with BotX API."""

    known_hosts: List[ExpressServer]

    def get_cts_by_host(self, host: str) -> ExpressServer:
        """Find CTS in bot registered servers.

        Arguments:
            host: host of server that should be found.

        Returns:
            Found instance of registered server.

        Raises:
            ServerUnknownError: raised if server was not found.
        """
        for cts in self.known_hosts:
            if cts.host == host:
                return cts

        raise ServerUnknownError(host=host)

    def get_token_for_cts(self, host: str) -> str:
        """Search token in bot saved tokens.

        Arguments:
            host: host for which token should be searched.

        Returns:
            Found token.

        Raises:
            ValueError: raised of there is not token for host.
        """
        server = self.get_cts_by_host(host)
        if server.server_credentials is not None:
            return server.server_credentials.token

        raise ValueError("token for cts {0} unfilled".format(host))



"""Endpoints for notification resource."""
from molten import RequestData, Response, Settings

from botx.clients.methods.base import APIResponse
from botx.clients.methods.v3.notification.direct_notification import NotificationDirect
from botx.clients.methods.v3.notification.notification import Notification
from botx.testing.botx_mock.binders import bind_implementation_to_method
from botx.testing.botx_mock.wsgi.messages import add_message_to_collection
from botx.testing.botx_mock.wsgi.responses import (
    PydanticResponse,
    generate_push_response,
)


@bind_implementation_to_method(Notification)
def post_notification(request_data: RequestData, settings: Settings) -> Response:
    """Handle pushed notification request.

    Arguments:
        request_data: parsed json data from request.
        settings: application settings with storage.

    Returns:
        Response with sync_id of pushed message.
    """
    payload = Notification.parse_obj(request_data)
    add_message_to_collection(settings, payload)
    return PydanticResponse(APIResponse[str](result="notification_pushed"))


@bind_implementation_to_method(NotificationDirect)
def post_notification_direct(request_data: RequestData, settings: Settings) -> Response:
    """Handle pushed notification request.

    Arguments:
        request_data: parsed json data from request.
        settings: application settings with storage.

    Returns:
        Response with sync_id of pushed message.
    """
    payload = NotificationDirect.parse_obj(request_data)
    add_message_to_collection(settings, payload)
    return generate_push_response(payload)



"""Aliases for complex types from `typing`."""

from typing import TYPE_CHECKING, Awaitable, Callable, TypeVar, Union

from botx.models.messages import message

if TYPE_CHECKING:
    from botx.bots.bots import Bot  # noqa: WPS433

try:
    from typing import Protocol, Literal  # noqa: WPS433
except ImportError:
    from typing_extensions import (  # type: ignore  # noqa: WPS433, WPS440, F401
        Protocol,
        Literal,
    )

ExceptionT = TypeVar("ExceptionT", bound=Exception)

# Something that can handle new message
AsyncExecutor = Callable[[message.Message], Awaitable[None]]
SyncExecutor = Callable[[message.Message], None]
Executor = Union[AsyncExecutor, SyncExecutor]

# Middlware dispatchers
AsyncMiddlewareDispatcher = Callable[[message.Message, AsyncExecutor], Awaitable[None]]
SyncMiddlewareDispatcher = Callable[[message.Message, SyncExecutor], None]
MiddlewareDispatcher = Union[AsyncMiddlewareDispatcher, SyncMiddlewareDispatcher]

# Exception handlers
AsyncExceptionHandler = Callable[[ExceptionT, message.Message], Awaitable[None]]
SyncExceptionHandler = Callable[[ExceptionT, message.Message], None]
ExceptionHandler = Union[AsyncExceptionHandler, SyncExceptionHandler]

# Startup and shutdown events
AsyncLifespanEvent = Callable[["Bot"], Awaitable[None]]
SyncLifespanEvent = Callable[["Bot"], None]
BotLifespanEvent = Union[AsyncLifespanEvent, SyncLifespanEvent]



"""Converters for common operations."""

from typing import List, Optional, Sequence, TypeVar

TSequenceElement = TypeVar("TSequenceElement")


def optional_sequence_to_list(
    seq: Optional[Sequence[TSequenceElement]] = None,
) -> List[TSequenceElement]:
    """Convert optional sequence of elements to list.

    Arguments:
        seq: sequence that should be converted to list.

    Returns:
        List of passed elements.
    """
    return list(seq or [])



"""Pydantic models for bot menu."""

from typing import List

from pydantic import BaseModel

from botx.models.enums import Statuses


class MenuCommand(BaseModel):
    """Command that is shown in bot menu."""

    #: command description that will be shown in menu.
    description: str

    #: command body that will trigger command execution.
    body: str

    #: command name.
    name: str


class StatusResult(BaseModel):
    """Bot menu commands collection."""

    #: is bot enabled.
    enabled: bool = True

    #: status of bot.
    status_message: str = "Bot is working"

    #: list of bot commands that will be shown in menu.
    commands: List[MenuCommand] = []


class Status(BaseModel):
    """Object that should be returned on `/status` request from BotX API."""

    #: operation status.
    status: Statuses = Statuses.ok

    #: bot status.
    result: StatusResult = StatusResult()



"""Endpoints for command resource."""

from starlette.requests import Request
from starlette.responses import Response

from botx.clients.methods.v3.command.command_result import CommandResult
from botx.testing.botx_mock.asgi.messages import add_message_to_collection
from botx.testing.botx_mock.asgi.responses import generate_push_response
from botx.testing.botx_mock.binders import bind_implementation_to_method


@bind_implementation_to_method(CommandResult)
async def post_command_result(request: Request) -> Response:
    """Handle command result request.

    Arguments:
        request: HTTP request from Starlette.

    Returns:
        Response with sync_id of pushed message.
    """
    payload = CommandResult.parse_obj(await request.json())
    add_message_to_collection(request, payload)
    return generate_push_response(payload)



"""Mixin for shortcut for command resource requests."""

from typing import cast
from uuid import UUID

from botx.bots.mixins.requests.call_protocol import BotXMethodCallProtocol
from botx.clients.methods.v3.command.command_result import CommandResult
from botx.clients.types.message_payload import ResultPayload
from botx.clients.types.options import ResultOptions
from botx.models.messages.sending.credentials import SendingCredentials
from botx.models.messages.sending.payload import MessagePayload


class CommandRequestsMixin:
    """Mixin for shortcut for command resource requests."""

    async def send_command_result(
        self: BotXMethodCallProtocol,
        credentials: SendingCredentials,
        payload: MessagePayload,
    ) -> UUID:
        """Send command result into chat.

        Arguments:
            credentials: credentials for making request.
            payload: payload for command result.

        Returns:
             ID sent message.
        """
        return await self.call_method(
            CommandResult(
                sync_id=cast(UUID, credentials.sync_id),
                event_sync_id=credentials.message_id,
                result=ResultPayload(
                    body=payload.text,
                    metadata=payload.metadata,
                    bubble=payload.markup.bubbles,
                    keyboard=payload.markup.keyboard,
                    mentions=payload.options.mentions,
                ),
                recipients=payload.options.recipients,
                file=payload.file,
                opts=ResultOptions(
                    silent_response=payload.options.silent_response,
                    notification_opts=payload.options.notifications,
                ),
            ),
            credentials=credentials,
        )



"""Special options for messages from bot."""
from pydantic import BaseModel

from botx.models.messages.sending.options import NotificationOptions


class ResultOptions(BaseModel):
    """Configuration for command result or notification that is send to BotX API."""

    #: send message only when stealth mode is enabled.
    stealth_mode: bool = False

    #: don't show next user messages in chat
    silent_response: bool = False

    #: message options for configuring notifications.
    notification_opts: NotificationOptions = NotificationOptions()



"""Special options for message."""

from typing import List

from pydantic import BaseModel

from botx.models.mentions import Mention
from botx.models.typing import AvailableRecipients


class NotificationOptions(BaseModel):
    """Configurations for message notifications."""

    #: show notification about message.
    send: bool = True

    #: break mute on bot messages.
    force_dnd: bool = False


class MessageOptions(BaseModel):
    """Message options configuration."""

    #: users that should receive message.
    recipients: AvailableRecipients = "all"

    #: attached to message mentions.
    mentions: List[Mention] = []

    #: don't show next user messages in chat
    silent_response: bool = False

    #: notification configuration.
    notifications: NotificationOptions = NotificationOptions()



"""Endpoints for chats resource."""
import uuid

from molten import Request, RequestData, Response, Settings

from botx.clients.methods.base import APIResponse
from botx.clients.methods.v3.chats import (
    add_admin_role,
    add_user,
    create,
    info,
    remove_user,
    stealth_disable,
    stealth_set,
)
from botx.clients.types.response_results import ChatCreatedResult
from botx.models import chats, enums, users
from botx.testing.botx_mock.binders import bind_implementation_to_method
from botx.testing.botx_mock.wsgi.messages import add_request_to_collection
from botx.testing.botx_mock.wsgi.responses import PydanticResponse


@bind_implementation_to_method(info.Info)
def get_info(request: Request, settings: Settings) -> Response:
    """Handle retrieving information of chat request.

    Arguments:
        request: HTTP request from Molten.
        settings: application settings with storage.

    Returns:
        Response with information of chat.
    """
    payload = info.Info.parse_obj(request.params)
    add_request_to_collection(settings, payload)
    return PydanticResponse(
        APIResponse[chats.ChatFromSearch](
            result=chats.ChatFromSearch(
                name="chat name",
                chat_type=enums.ChatTypes.group_chat,
                creator=uuid.uuid4(),
                group_chat_id=payload.group_chat_id,
                members=[
                    users.UserFromChatSearch(
                        user_huid=uuid.uuid4(),
                        user_kind=enums.UserKinds.user,
                        admin=True,
                    ),
                ],
            ),
        ),
    )


@bind_implementation_to_method(add_user.AddUser)
def post_add_user(request_data: RequestData, settings: Settings) -> Response:
    """Handle adding of user to chat request.

    Arguments:
        request_data: parsed json data from request.
        settings: application settings with storage.

    Returns:
        Response with result of adding.
    """
    payload = add_user.AddUser.parse_obj(request_data)
    add_request_to_collection(settings, payload)
    return PydanticResponse(APIResponse[bool](result=True))


@bind_implementation_to_method(remove_user.RemoveUser)
def post_remove_user(request_data: RequestData, settings: Settings) -> Response:
    """Handle removing of user to chat request.

    Arguments:
        request_data: parsed json data from request.
        settings: application settings with storage.

    Returns:
        Response with result of removing.
    """
    payload = remove_user.RemoveUser.parse_obj(request_data)
    add_request_to_collection(settings, payload)
    return PydanticResponse(APIResponse[bool](result=True))


@bind_implementation_to_method(stealth_set.StealthSet)
def post_stealth_set(request_data: RequestData, settings: Settings) -> Response:
    """Handle stealth enabling in chat request.

    Arguments:
        request_data: parsed json data from request.
        settings: application settings with storage.

    Returns:
        Response with result of enabling stealth.
    """
    payload = stealth_set.StealthSet.parse_obj(request_data)
    add_request_to_collection(settings, payload)
    return PydanticResponse(APIResponse[bool](result=True))


@bind_implementation_to_method(stealth_disable.StealthDisable)
def post_stealth_disable(request_data: RequestData, settings: Settings) -> Response:
    """Handle stealth disabling in chat request.

    Arguments:
        request_data: parsed json data from request.
        settings: application settings with storage.

    Returns:
        Response with result of disabling stealth.
    """
    payload = stealth_disable.StealthDisable.parse_obj(request_data)
    add_request_to_collection(settings, payload)
    return PydanticResponse(APIResponse[bool](result=True))


@bind_implementation_to_method(create.Create)
def post_create(request_data: RequestData, settings: Settings) -> Response:
    """Handle creation of new chat request.

    Arguments:
        request_data: parsed json data from request.
        settings: application settings with storage.

    Returns:
        Response with result of creation.
    """
    payload = create.Create.parse_obj(request_data)
    add_request_to_collection(settings, payload)
    return PydanticResponse(
        APIResponse[ChatCreatedResult](result=ChatCreatedResult(chat_id=uuid.uuid4())),
    )


@bind_implementation_to_method(add_admin_role.AddAdminRole)
def post_add_admin_role(request_data: RequestData, settings: Settings) -> Response:
    """Handle promoting users to admins request.

    Arguments:
        request_data: parsed json data from request.
        settings: application settings with storage.

    Returns:
        Response with result of adding.
    """
    payload = add_admin_role.AddAdminRole.parse_obj(request_data)
    add_request_to_collection(settings, payload)
    return PydanticResponse(APIResponse[bool](result=True))



"""Common responses for mocks."""

import uuid
from typing import Any, Dict, Optional, Union

from molten import HTTP_200, Response
from pydantic import BaseModel

from botx.clients.methods.base import APIResponse
from botx.clients.methods.v3.command.command_result import CommandResult
from botx.clients.methods.v3.notification.direct_notification import NotificationDirect
from botx.clients.types.response_results import PushResult


class PydanticResponse(Response):
    """Custom response to encode pydantic model from route."""

    def __init__(
        self,
        model: BaseModel,
        status_code: str = HTTP_200,
        headers: Optional[Dict[Any, Any]] = None,
    ) -> None:
        """Init custom response.

        Arguments:
            model: pydantic model that should be encoded.
            status_code: response HTTP status code.
            headers: headers for response.
        """
        headers = headers or {}
        headers["Content-Type"] = "application/json"
        super().__init__(
            status_code, headers, model.json(by_alias=True),
        )


def generate_push_response(
    payload: Union[CommandResult, NotificationDirect],
) -> Response:
    """Generate response as like new message from bot was pushed.

    Arguments:
        payload: pushed message.

    Returns:
        Response with sync_id for new message.
    """
    sync_id = payload.event_sync_id or uuid.uuid4()
    return PydanticResponse(
        APIResponse[PushResult](result=PushResult(sync_id=sync_id)),
    )



"""Definition of base for custom middlewares.

Important:
    Middleware should implement `dispatch` method that can be a common function or
    an asynchronous function.

```python
class MyAsyncBotXMiddleware(BaseMiddleware):
    async def dispatch(
        self, message: Message, call_next: AsyncExecutor,
    ) -> None:
        await call_next(message)

class MySyncBotXMiddleware(BaseMiddleware):
    def dispatch(self, message: Message, call_next: SyncExecutor) -> None:
        call_next(message)
```
"""

from typing import Callable, Optional

from botx import concurrency
from botx.models.messages.message import Message
from botx.typing import Executor, MiddlewareDispatcher, SyncExecutor


def _default_dispatch(
    _middleware: "BaseMiddleware", _message: Message, _call_next: SyncExecutor,
) -> None:
    raise NotImplementedError


class BaseMiddleware:
    """Base middleware entity."""

    dispatch: Callable = _default_dispatch

    def __init__(
        self, executor: Executor, dispatch: Optional[MiddlewareDispatcher] = None,
    ) -> None:
        """Init middleware with required query_params.

        Arguments:
            executor: callable object that accept message and will be executed after
                middlewares.
            dispatch: middleware logic executor.
        """
        self.executor = executor
        self.dispatch_func = dispatch or self.dispatch

    async def __call__(self, message: Message) -> None:
        """Call middleware dispatcher as normal handler executor.

        Arguments:
            message: incoming message.
        """
        executor = self.executor
        if not concurrency.is_awaitable(self.dispatch_func):
            executor = concurrency.async_to_sync(self.executor)

        await concurrency.callable_to_coroutine(self.dispatch_func, message, executor)



"""Mixin that defines handler decorator."""

from typing import Any, Callable, Optional, Sequence, Union

from botx.collecting.collectors.collector import Collector
from botx.dependencies.models import Depends


class DefaultHandlerMixin:
    """Mixin that defines handler decorator."""

    collector: Collector

    def default(  # noqa: WPS211
        self,
        handler: Optional[Callable] = None,
        *,
        command: Optional[str] = None,
        commands: Optional[Sequence[str]] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        full_description: Optional[str] = None,
        include_in_status: Union[bool, Callable] = False,
        dependencies: Optional[Sequence[Depends]] = None,
        dependency_overrides_provider: Any = None,
    ) -> Callable:
        """Add new handler to bot and register it as default handler.

        !!! info
            If `include_in_status` is a function, then `body` argument will be checked
            for matching public commands style, like `/command`.

        Arguments:
            handler: callable that will be used for executing handler.
            command: body template that will trigger this handler.
            commands: list of body templates that will trigger this handler.
            name: optional name for handler that will be used in generating body.
            description: description for command that will be shown in bot's menu.
            full_description: full description that can be used for example in `/help`
                command.
            include_in_status: should this handler be shown in bot's menu, can be
                callable function with no arguments *(for now)*.
            dependencies: sequence of dependencies that should be executed before
                handler.
            dependency_overrides_provider: mock of callable for handler.

        Returns:
            Passed in `handler` callable.
        """
        return self.collector.default(
            handler=handler,
            command=command,
            commands=commands,
            name=name,
            description=description,
            full_description=full_description,
            include_in_status=include_in_status,
            dependencies=dependencies,
            dependency_overrides_provider=dependency_overrides_provider,
        )



"""Predefined mocked entities builder for routes."""
import uuid
from typing import Optional, Tuple

from botx.models.users import UserFromSearch


def create_test_user(
    *,
    user_huid: Optional[uuid.UUID] = None,
    email: Optional[str] = None,
    ad: Optional[Tuple[str, str]] = None,
) -> UserFromSearch:
    """Build test user for using in search.

    Arguments:
        user_huid: HUID of user for search.
        email: email of user for search.
        ad: AD credentials of user for search.

    Returns:
        "Found" user.
    """
    return UserFromSearch(
        user_huid=user_huid or uuid.uuid4(),
        ad_login=ad[0] if ad else "ad_login",
        ad_domain=ad[1] if ad else "ad_domain",
        name="test user",
        company="test company",
        company_position="test position",
        department="test department",
        emails=[email or "test@example.com"],
    )



"""Definition for "chat creation error"."""
from typing import NoReturn

from httpx import Response

from botx.clients.methods.base import APIErrorResponse, BotXMethod
from botx.exceptions import BotXAPIError


class ChatCreationError(BotXAPIError):
    """Error for raising when there is error for chat creation."""

    message_template = "error while creating chat"


def handle_error(method: BotXMethod, response: Response) -> NoReturn:
    """Handle "chat creation error" error response.

    Arguments:
        method: method which was made before error.
        response: HTTP response from BotX API.

    Raises:
        ChatCreationError: raised always.
    """
    APIErrorResponse[dict].parse_obj(response.json())
    raise ChatCreationError(
        url=method.url,
        method=method.http_method,
        response_content=response.content,
        status_content=response.status_code,
    )



"""Definition for command handler."""

from __future__ import annotations

import re
from dataclasses import field
from typing import Any, Callable, List, Optional, Union

from pydantic import validator
from pydantic.dataclasses import dataclass

from botx.collecting.handlers.validators import (
    check_handler_is_function,
    retrieve_dependant,
    retrieve_executor,
    retrieve_full_description_for_handler,
    retrieve_name_for_handler,
    validate_body_for_status,
)
from botx.dependencies import models as deps
from botx.models.messages.message import Message


@dataclass
class Handler:
    """Handler that will store body and callable."""

    #: callable for executing registered logic.
    handler: Callable

    #: command body.
    body: str

    #: name of handler.
    name: str = ""

    #: description that will be shown in bot's menu.
    description: Optional[str] = None

    #: custom description that can be used for another purposes.
    full_description: str = ""

    #: should handler be included into status.
    include_in_status: Union[bool, Callable] = True

    #: wrapper around handler that will be executed.
    dependant: deps.Dependant = field(init=False)

    #: background dependencies for handler.
    dependencies: List[deps.Depends] = field(default_factory=list)

    #: custom object that will override dependencies for handler.
    dependency_overrides_provider: Optional[Any] = None

    #: function that will be used for handling incoming message
    executor: Callable = field(init=False)

    _body_validator = validator("executor", always=True)(validate_body_for_status)
    _handler_is_function_validator = validator("handler", pre=True, always=True)(
        check_handler_is_function,
    )

    async def __call__(self, message: Message) -> None:
        """Execute handler using incoming message.

        Arguments:
            message: message that will be handled by handler.
        """
        await self.executor(message)

    def __post_init__(self) -> None:
        """Initialize or update special fields."""
        self.name = retrieve_name_for_handler(self.name, self.handler)
        self.full_description = retrieve_full_description_for_handler(
            self.full_description, self.handler,
        )
        self.dependant = retrieve_dependant(self.handler, self.dependencies)
        self.executor = retrieve_executor(  # type: ignore
            self.dependant, self.dependency_overrides_provider,
        )

    def matches(self, message: Message) -> bool:
        """Check if message body matched to handler's body.

        Arguments:
            message: incoming message which body will be used to check route.

        Returns:
            Result of check.
        """
        return bool(re.compile(self.body).match(message.body))

    def command_for(self, *args: Any) -> str:
        """Build a command string using passed body query_params.

        Arguments:
            args: sequence of elements that are arguments for command.

        Returns:
            Built command.
        """
        args_str = " ".join((str(arg) for arg in args[1:]))
        return "{0} {1}".format(self.body, args_str).strip()

    def __eq__(self, other: object) -> bool:
        """Compare 2 handlers for equality.

        Arguments:
            other: handler to compare with.

        Returns:
            Result of comparing.
        """
        if not isinstance(other, Handler):
            return False

        callable_comp = self.handler == other.handler
        callable_comp = callable_comp and self.dependencies == other.dependencies

        export_comp = self.name == other.name
        export_comp = export_comp and self.body == other.body
        export_comp = export_comp and self.description == other.description
        export_comp = export_comp and self.full_description == other.full_description
        export_comp = export_comp and self.include_in_status == other.include_in_status

        return callable_comp and export_comp



"""Definition for sending credentials."""

from typing import Optional
from uuid import UUID

from pydantic import BaseModel


class SendingCredentials(BaseModel):
    """Credentials that are required to send command or notification result."""

    #: message event id.
    sync_id: Optional[UUID] = None

    #: id of message that will be sent.
    message_id: Optional[UUID] = None

    #: chat id in which bot should send message.
    chat_id: Optional[UUID] = None

    #: bot that handles message.
    bot_id: Optional[UUID] = None

    #: host on which bot answers.
    host: Optional[str] = None

    #: token that is used for bot authorization on requests to BotX API.
    token: Optional[str] = None



"""Exception mixin for bot."""

from typing import Callable, Type

from botx import typing
from botx.middlewares.exceptions import ExceptionMiddleware

try:
    from typing import Protocol  # noqa: WPS433
except ImportError:
    from typing_extensions import (  # type: ignore  # noqa: WPS433, WPS440, F401
        Protocol,
    )


class ExceptionHandlersMixin:
    """Mixin that defines functions for exception handlers registration."""

    exception_middleware: ExceptionMiddleware

    def add_exception_handler(
        self, exc_class: Type[Exception], handler: typing.ExceptionHandler,
    ) -> None:
        """Register new handler for exception.

        Arguments:
            exc_class: exception type that should be handled.
            handler: handler for exception.
        """
        self.exception_middleware.add_exception_handler(exc_class, handler)

    def exception_handler(self, exc_class: Type[Exception]) -> Callable:
        """Register callable as handler for exception.

        Arguments:
            exc_class: exception type that should be handled.

        Returns:
            Decorator that will register exception and return passed function.
        """

        def decorator(handler: typing.ExceptionHandler) -> Callable:
            self.add_exception_handler(exc_class, handler)
            return handler

        return decorator



"""Functions for solving dependencies."""

from typing import Any, Awaitable, Callable, Dict, Optional, Tuple, cast

from botx import concurrency
from botx.dependencies.models import (
    CacheKey,
    Dependant,
    DependenciesCache,
    get_dependant,
)
from botx.models.messages.message import Message


async def solve_sub_dependency(
    message: Message,
    dependant: Dependant,
    solved_values: Dict[str, Any],
    dependency_overrides_provider: Any,
    dependency_cache: Dict[CacheKey, Any],
) -> None:
    """
    Solve single sub dependency.

    Arguments:
        message: incoming message that is used for solving this sub dependency.
        dependant: dependency that is solving while calling this function.
        solved_values: already filled validated_values that are required for this
            dependency.
        dependency_overrides_provider: an object with `dependency_overrides` attribute
            that contains overrides for dependencies.
        dependency_cache: cache that contains already solved dependency and result for
            it.

    """
    call = cast(Callable, dependant.call)
    use_sub_dependant = dependant

    overrides = getattr(
        dependency_overrides_provider
        if dependency_overrides_provider is not None
        else message.bot,
        "dependency_overrides",
        {},
    )
    if overrides:
        call = overrides.get(dependant.call, dependant.call)
        use_sub_dependant = get_dependant(call=call, name=dependant.name)

    solving_result = await solve_dependencies(
        message=message,
        dependant=use_sub_dependant,
        dependency_overrides_provider=dependency_overrides_provider,
        dependency_cache=dependency_cache,
    )
    dependency_cache.update(solving_result[1])

    dependant.cache_key = dependant.cache_key
    if dependant.use_cache and dependant.cache_key in dependency_cache:
        solved = dependency_cache[dependant.cache_key]
    else:
        solved = await concurrency.callable_to_coroutine(call, **solving_result[0])

    if dependant.name is not None:
        solved_values[dependant.name] = solved
    if dependant.cache_key not in dependency_cache:
        dependency_cache[dependant.cache_key] = solved


async def solve_dependencies(
    *,
    message: Message,
    dependant: Dependant,
    dependency_overrides_provider: Any = None,
    dependency_cache: Optional[Dict[CacheKey, Any]] = None,
) -> Tuple[Dict[str, Any], DependenciesCache]:
    """
    Resolve all required dependencies for Dependant using incoming message.

    Arguments:
        message: incoming Message with all necessary data.
        dependant: Dependant object for which all sub dependencies should be solved.
        dependency_overrides_provider: an object with `dependency_overrides` attribute
            that contains overrides for dependencies.
        dependency_cache: cache that contains already solved dependency and result for
            it.

    Returns:
        Keyword arguments with their vales and cache.

    """
    solved_values: Dict[str, Any] = {}
    dependency_cache = dependency_cache or {}
    for sub_dependant in dependant.dependencies:
        await solve_sub_dependency(
            message=message,
            dependant=sub_dependant,
            solved_values=solved_values,
            dependency_overrides_provider=dependency_overrides_provider,
            dependency_cache=dependency_cache,
        )

    if dependant.message_param_name:
        solved_values[dependant.message_param_name] = message
    if dependant.bot_param_name:
        solved_values[dependant.bot_param_name] = message.bot
    if dependant.async_client_param_name:
        solved_values[dependant.async_client_param_name] = message.bot.client
    if dependant.sync_client_param_name:
        solved_values[dependant.sync_client_param_name] = message.bot.sync_client
    return solved_values, dependency_cache


def get_executor(
    dependant: Dependant, dependency_overrides_provider: Any = None,
) -> Callable[[Message], Awaitable[None]]:
    """Get an execution callable for passed dependency.

    Arguments:
        dependant: passed dependency for which execution callable should be generated.
        dependency_overrides_provider: dependency overrider that will be passed to the
            execution.

    Returns:
        Asynchronous executor for handling message.

    Raises:
        AssertionError: raised if there is no callable in `dependant.call`.
    """
    if dependant.call is None:
        raise AssertionError("dependant.call must be present")

    async def factory(message: Message) -> None:
        solved_values, _ = await solve_dependencies(
            message=message,
            dependant=dependant,
            dependency_overrides_provider=dependency_overrides_provider,
        )
        await concurrency.callable_to_coroutine(
            cast(Callable, dependant.call), **solved_values,
        )

    return factory



"""Method for sending notification into many chats."""
from typing import List, Optional
from uuid import UUID

from pydantic import Field

from botx.clients.methods.base import AuthorizedBotXMethod
from botx.clients.types.message_payload import ResultPayload
from botx.clients.types.options import ResultOptions
from botx.models.files import File
from botx.models.typing import AvailableRecipients


class Notification(AuthorizedBotXMethod[str]):
    """Method for sending notification into many chats."""

    __url__ = "/api/v3/botx/notification/callback"
    __method__ = "POST"
    __returning__ = str

    #: IDs of chats for new notification.
    group_chat_ids: List[UUID] = []

    #: HUIDs of users that should receive notifications.
    recipients: AvailableRecipients = "all"

    #: data for build message: body, markup, mentions.
    result: ResultPayload = Field(..., alias="notification")

    #: attached file for message.
    file: Optional[File] = None

    #: extra options for message.
    opts: ResultOptions = ResultOptions()



"""Method for retrieving token for bot."""
from typing import Dict
from uuid import UUID

from httpx import URL

from botx.clients.methods.base import BotXMethod, PrimitiveDataType


class Token(BotXMethod[str]):
    """Method for retrieving token for bot."""

    __url__ = "/api/v2/botx/bots/{bot_id}/token"
    __method__ = "GET"
    __returning__ = str

    #: ID of bot which access for token.
    bot_id: UUID

    #: calculated signature from secret_key for bot.
    signature: str

    @property
    def url(self) -> str:
        """Full URL for request with filling bot_id."""
        super_url = URL(super().url)
        api_url = self.__url__.format(bot_id=self.bot_id)
        return str(super_url.join(api_url))

    @property
    def query_params(self) -> Dict[str, PrimitiveDataType]:
        """Query string query_params for request with signature key."""
        return {"signature": self.signature}

    def encode(self) -> None:
        """Return nothing to override request body.

        Returns:
            Nothing.
        """
        return None  # noqa: WPS324



"""Endpoints for events resource."""
from molten import RequestData, Response, Settings

from botx.clients.methods.base import APIResponse
from botx.clients.methods.v3.events.edit_event import EditEvent
from botx.testing.botx_mock.binders import bind_implementation_to_method
from botx.testing.botx_mock.wsgi.messages import add_message_to_collection
from botx.testing.botx_mock.wsgi.responses import PydanticResponse


@bind_implementation_to_method(EditEvent)
def post_edit_event(request_data: RequestData, settings: Settings) -> Response:
    """Handle edition of event request.

    Arguments:
        request_data: parsed json data from request.
        settings: application settings with storage.

    Returns:
        Empty json response.
    """
    payload = EditEvent.parse_obj(request_data)
    add_message_to_collection(settings, payload)
    return PydanticResponse(APIResponse[str](result="update_pushed"))



"""Definition of errors in processing request from BotX API."""

from typing import Any, Dict, List, Union

from pydantic import BaseModel, validator


class BotDisabledErrorData(BaseModel):
    """Data about occurred error."""

    #: message that will be shown to user.
    status_message: str


class BotDisabledResponse(BaseModel):
    """Response to BotX API if there was an error in handling incoming request."""

    #: error reason. *This should always be `bot_disabled` string.*
    reason: str = "bot_disabled"

    #: data about occurred error that should include `status_message` field in json.
    error_data: Union[Dict[str, Any], BotDisabledErrorData]

    errors: List[str] = []

    @validator("error_data", always=True, whole=True)
    def status_message_in_error_data(
        cls, error_data: Dict[str, Any],  # noqa: N805
    ) -> Union[BotDisabledErrorData, Dict[str, Any]]:
        """Check that value contains `status_message` key or field.

        Arguments:
            cls: this class.
            error_data: value that should be checked.

        Returns:
            Built payload for response.

        Raises:
            ValueError: raised if error_data does not contain status_message.
        """
        if set(error_data.keys()) == {"status_message"}:
            return BotDisabledErrorData(status_message=error_data["status_message"])

        if "status_message" not in error_data:
            raise ValueError("status_message key required in error_data")

        return error_data



"""Implementation for bot classes."""

import asyncio
from dataclasses import InitVar, field
from typing import Any, Callable, Dict, List, Set

from loguru import logger
from pydantic.dataclasses import dataclass

from botx import concurrency, dataclasses_config, exception_handlers, exceptions, typing
from botx.bots.mixins import (
    clients,
    collectors,
    exceptions as exception_mixin,
    lifespan,
    middlewares,
)
from botx.clients.clients import async_client, sync_client as synchronous_client
from botx.collecting.collectors.collector import Collector
from botx.dependencies.models import Depends
from botx.middlewares.exceptions import ExceptionMiddleware
from botx.models import credentials, datastructures, menu
from botx.models.messages.message import Message


@dataclass(config=dataclasses_config.BotXDataclassConfig)
class Bot(  # noqa: WPS215
    collectors.BotCollectingMixin,
    clients.ClientsMixin,
    lifespan.LifespanMixin,
    middlewares.MiddlewareMixin,
    exception_mixin.ExceptionHandlersMixin,
):
    """Class that implements bot behaviour."""

    dependencies: InitVar[List[Depends]] = field(default=None)
    known_hosts: List[credentials.ExpressServer] = field(default_factory=list)
    startup_events: List[typing.BotLifespanEvent] = field(default_factory=list)
    shutdown_events: List[typing.BotLifespanEvent] = field(default_factory=list)

    client: async_client.AsyncClient = field(init=False)
    sync_client: synchronous_client.Client = field(init=False)
    collector: Collector = field(init=False)
    exception_middleware: ExceptionMiddleware = field(init=False)
    state: datastructures.State = field(init=False)
    dependency_overrides: Dict[Callable, Callable] = field(
        init=False, default_factory=dict,
    )

    tasks: Set[asyncio.Future] = field(init=False, default_factory=set)

    async def __call__(self, message: Message) -> None:
        """Iterate through collector, find handler and execute it, running middlewares.

        Arguments:
            message: message that will be proceed by handler.
        """
        self.tasks.add(asyncio.ensure_future(self.exception_middleware(message)))

    def __post_init__(self, dependencies: List[Depends]) -> None:
        """Initialize special fields.

        Arguments:
            dependencies: initial background dependencies for inner collector.
        """
        self.state = datastructures.State()
        self.client = async_client.AsyncClient()
        self.sync_client = synchronous_client.Client()
        self.collector = Collector(
            dependencies=dependencies, dependency_overrides_provider=self,
        )
        self.exception_middleware = ExceptionMiddleware(self.collector)

        self.add_exception_handler(
            exceptions.DependencyFailure,
            exception_handlers.dependency_failure_exception_handler,
        )
        self.add_exception_handler(
            exceptions.NoMatchFound,
            exception_handlers.no_match_found_exception_handler,
        )

    async def status(self, *args: Any, **kwargs: Any) -> menu.Status:
        """Generate status object that could be return to BotX API on `/status`.

        Arguments:
            args: additional positional arguments that will be passed to callable
                status function.
            kwargs: additional key arguments that will be passed to callable
                status function.

        Returns:
            Built status for returning to BotX API.
        """
        status = menu.Status()
        for handler in self.handlers:
            if callable(handler.include_in_status):
                include_in_status = await concurrency.callable_to_coroutine(
                    handler.include_in_status, *args, **kwargs,
                )
            else:
                include_in_status = handler.include_in_status

            if include_in_status:
                status.result.commands.append(
                    menu.MenuCommand(
                        description=handler.description or "",
                        body=handler.body,
                        name=handler.name,
                    ),
                )

        return status

    async def execute_command(self, message: dict) -> None:
        """Process data with incoming message and handle command inside.

        Arguments:
            message: incoming message to bot.

        Raises:
            ServerUnknownError: raised if message was received from unregistered host.
        """
        logger.bind(botx_bot=True, payload=message).debug("process incoming message")
        msg = Message.from_dict(message, self)
        for server in self.known_hosts:
            if server.host == msg.host:
                await self(msg)
                break
        else:
            raise exceptions.ServerUnknownError(host=msg.host)



"""Endpoints for events resource."""

from starlette.requests import Request
from starlette.responses import Response

from botx.clients.methods.base import APIResponse
from botx.clients.methods.v3.events.edit_event import EditEvent
from botx.testing.botx_mock.asgi.messages import add_message_to_collection
from botx.testing.botx_mock.asgi.responses import PydanticResponse
from botx.testing.botx_mock.binders import bind_implementation_to_method


@bind_implementation_to_method(EditEvent)
async def post_edit_event(request: Request) -> Response:
    """Handle edition of event request.

    Arguments:
        request: HTTP request from Starlette.

    Returns:
        Empty json response.
    """
    payload = EditEvent.parse_obj(await request.json())
    add_message_to_collection(request, payload)
    return PydanticResponse(APIResponse[str](result="update_pushed"))



"""Mixin for shortcut for chats resource requests."""

from typing import List, Optional
from uuid import UUID

from botx.bots.mixins.requests.call_protocol import BotXMethodCallProtocol
from botx.clients.methods.v3.chats.add_admin_role import AddAdminRole
from botx.clients.methods.v3.chats.add_user import AddUser
from botx.clients.methods.v3.chats.create import Create
from botx.clients.methods.v3.chats.info import Info
from botx.clients.methods.v3.chats.remove_user import RemoveUser
from botx.clients.methods.v3.chats.stealth_disable import StealthDisable
from botx.clients.methods.v3.chats.stealth_set import StealthSet
from botx.models.chats import ChatFromSearch
from botx.models.enums import ChatTypes
from botx.models.messages.sending.credentials import SendingCredentials


class ChatsRequestsMixin:
    """Mixin for shortcut for chats resource requests."""

    async def create_chat(  # noqa: WPS211
        self: BotXMethodCallProtocol,
        credentials: SendingCredentials,
        name: str,
        members: List[UUID],
        chat_type: ChatTypes,
        description: Optional[str] = None,
        avatar: Optional[str] = None,
    ) -> UUID:
        """Create new chat.

        Arguments:
            credentials: credentials for making request.
            name: name of chat that should be created.
            members: HUIDs of users that should be added into chat.
            chat_type: chat type.
            description: description of new chat.
            avatar: logo image of chat.

        Returns:
             ID of created chat.
        """
        return await self.call_method(
            Create(
                name=name,
                description=description,
                members=members,
                avatar=avatar,
                chat_type=chat_type,
            ),
            credentials=credentials,
        )

    async def get_chat_info(
        self: BotXMethodCallProtocol, credentials: SendingCredentials, chat_id: UUID,
    ) -> ChatFromSearch:
        """Create new chat.

        Arguments:
            credentials: credentials for making request.
            chat_id: ID of chat for about which information should be retrieving.

        Returns:
             Information about chat.
        """
        return await self.call_method(
            Info(group_chat_id=chat_id), credentials=credentials,
        )

    async def enable_stealth_mode(  # noqa: WPS211
        self: BotXMethodCallProtocol,
        credentials: SendingCredentials,
        chat_id: UUID,
        disable_web: bool = False,
        burn_in: Optional[int] = None,
        expire_in: Optional[int] = None,
    ) -> None:
        """Enable stealth mode.

        Arguments:
            credentials: credentials for making request.
            chat_id: ID of chat where stealth should be enabled.
            disable_web: should messages be shown in web.
            burn_in: time of messages burning after read.
            expire_in: time of messages burning after send.
        """
        await self.call_method(
            StealthSet(
                group_chat_id=chat_id,
                disable_web=disable_web,
                burn_in=burn_in,
                expire_in=expire_in,
            ),
            credentials=credentials,
        )

    async def disable_stealth_mode(
        self: BotXMethodCallProtocol, credentials: SendingCredentials, chat_id: UUID,
    ) -> None:
        """Disable stealth mode.

        Arguments:
            credentials: credentials for making request.
            chat_id: ID of chat where stealth should be disabled.
        """
        await self.call_method(
            StealthDisable(group_chat_id=chat_id), credentials=credentials,
        )

    async def add_users(
        self: BotXMethodCallProtocol,
        credentials: SendingCredentials,
        chat_id: UUID,
        user_huids: List[UUID],
    ) -> None:
        """Add users to chat.

        Arguments:
            credentials: credentials for making request.
            chat_id: ID of chat into which users should be added.
            user_huids: IDs of users that should be added into chat.
        """
        await self.call_method(
            AddUser(group_chat_id=chat_id, user_huids=user_huids),
            credentials=credentials,
        )

    async def remove_users(
        self: BotXMethodCallProtocol,
        credentials: SendingCredentials,
        chat_id: UUID,
        user_huids: List[UUID],
    ) -> None:
        """Remove users from chat.

        Arguments:
            credentials: credentials for making request.
            chat_id: ID of chat from which users should be removed.
            user_huids: HUID of users that should be removed.
        """
        await self.call_method(
            RemoveUser(group_chat_id=chat_id, user_huids=user_huids),
            credentials=credentials,
        )

    async def add_admin_roles(
        self: BotXMethodCallProtocol,
        credentials: SendingCredentials,
        chat_id: UUID,
        user_huids: List[UUID],
    ) -> None:
        """Promote users in chat to admins.

        Arguments:
            credentials: credentials for making request.
            chat_id: UUID of chat where action should be performed.
            user_huids: HUIDs of users that should be promoted to admins.
        """
        await self.call_method(
            AddAdminRole(group_chat_id=chat_id, user_huids=user_huids),
            credentials=credentials,
        )



"""Decorator for binding custom BotXMethod with implementation."""
from typing import Any, Callable, Type

from botx.clients.methods.base import BotXMethod


def bind_implementation_to_method(method: Type[BotXMethod]) -> Callable[..., Any]:
    """Bind implementation of async route to method.

    Arguments:
        method: method class to bind.

    Returns:
        Decorator that binds method and returns it.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        func.method = method  # type: ignore
        return func

    return decorator



"""Mixin for shortcut for bots resource requests."""

from uuid import UUID

from botx.bots.mixins.requests.call_protocol import BotXMethodCallProtocol
from botx.clients.methods.v2.bots.token import Token


class BotsRequestsMixin:
    """Mixin for shortcut for bots resource requests."""

    async def get_token(
        self: BotXMethodCallProtocol, host: str, bot_id: UUID, signature: str,
    ) -> str:
        """Obtain token for bot.

        Arguments:
            host: host on which request should be made.
            bot_id: ID of bot for which token should be obtained.
            signature: calculated signature of bot.

        Returns:
             Obtained token.
        """
        return await self.call_method(
            Token(bot_id=bot_id, signature=signature), host=host,
        )



"""Aliases for complex types from `typing` for models."""

from typing import List, Union
from uuid import UUID

from botx.models.buttons import BubbleElement, KeyboardElement

try:
    from typing import Literal  # noqa: WPS433
except ImportError:
    from typing_extensions import (  # type: ignore  # noqa: WPS433, WPS440, F401
        Literal,
    )

BubblesRow = List[BubbleElement]
BubbleMarkup = List[BubblesRow]

KeyboardRow = List[KeyboardElement]
KeyboardMarkup = List[KeyboardRow]

AvailableRecipients = Union[List[UUID], Literal["all"]]



"""Method for retrieving information about chat."""
from uuid import UUID

import httpx

from botx.clients.methods.base import AuthorizedBotXMethod
from botx.clients.methods.errors import messaging
from botx.models.chats import ChatFromSearch


class Info(AuthorizedBotXMethod[ChatFromSearch]):
    """Method for retrieving information about chat."""

    __url__ = "/api/v3/botx/chats/info"
    __method__ = "GET"
    __returning__ = ChatFromSearch
    __errors_handlers__ = {httpx.codes.BAD_REQUEST: messaging.handle_error}

    #: ID of chat for about which information should be retrieving.
    group_chat_id: UUID



"""Entities that represent some structs that are used in this library."""

from typing import Any, Optional


class State:
    """An object that can be used to store arbitrary state."""

    _state: dict

    def __init__(self, state: Optional[dict] = None):
        """Init state with required query_params.

        Arguments:
            state: initial state.
        """
        state = state or {}
        super().__setattr__("_state", state)  # noqa: WPS613

    def __setattr__(self, key: Any, new_value: Any) -> None:
        """Set state attribute.

        Arguments:
            key: key to set attribute.
            new_value: value of attribute.
        """
        self._state[key] = new_value

    def __getattr__(self, key: Any) -> Any:
        """Get state attribute.

        Arguments:
            key: key of retrieved attribute.

        Returns:
            Stored value.

        Raises:
            AttributeError: raised if attribute was not found in state.
        """
        try:
            return self._state[key]
        except KeyError:
            raise AttributeError("state has no attribute '{0}'".format(key))



"""Exceptions that are used in this library."""
from typing import Any, Dict


class BotXException(Exception):
    """Base error for exception in this library."""

    #: template that should be rendered on __str__ call.
    message_template: str = ""

    def __init__(self, **kwargs: Any) -> None:
        """Init exception with passed query_params.

        Arguments:
            kwargs: key-arguments that will be stored in instance.
        """
        self.__dict__ = kwargs

    def __str__(self) -> str:
        """Render string representation.

        Returns:
            String representation of error.
        """
        return self.message_template.format(**self.__dict__)


class NoMatchFound(BotXException):
    """Raised by collector if no matching handler exists."""

    message_template = "handler for {search_param} not found"

    #: body for which handler was not found.
    search_param: str


class DependencyFailure(BotXException):
    """Raised when there is error in dependency and flow should be stopped."""


class BotXAPIError(BotXException):
    """Raised if there is an error in requests to BotX API."""

    message_template = "unable to send {method} {url} to BotX API ({status})"

    #: URL from request.
    url: str

    #: HTTP method.
    method: str

    #: response from API.
    response_content: Dict[str, Any]

    # HTTP status code.
    status: int


class ServerUnknownError(BotXException):
    """Raised if bot does not know host."""

    message_template = "unknown server {host}"

    #: host that is unregistered.
    host: str



"""Definition for bot's collecting component.

All of this methods are just wrappers around inner collector.
"""

from typing import Any, List, Optional, Sequence

from botx.bots.mixins.collecting.add_handler import AddHandlerMixin
from botx.bots.mixins.collecting.default import DefaultHandlerMixin
from botx.bots.mixins.collecting.handler import HandlerMixin
from botx.bots.mixins.collecting.hidden import HiddenHandlerMixin
from botx.bots.mixins.collecting.system_events import SystemEventsHandlerMixin
from botx.collecting.collectors.collector import Collector
from botx.collecting.handlers.handler import Handler
from botx.dependencies import models as deps


class BotCollectingMixin(  # noqa: WPS215
    AddHandlerMixin,
    HandlerMixin,
    DefaultHandlerMixin,
    HiddenHandlerMixin,
    SystemEventsHandlerMixin,
):
    """Mixin that defines collector-like behaviour."""

    collector: Collector

    @property
    def handlers(self) -> List[Handler]:
        """Get handlers registered on this bot.

        Returns:
            Registered handlers of bot.
        """
        return self.collector.handlers

    def include_collector(
        self,
        collector: Collector,
        *,
        dependencies: Optional[Sequence[deps.Depends]] = None,
    ) -> None:
        """Include handlers from collector into bot.

        Arguments:
            collector: collector from which handlers should be copied.
            dependencies: optional sequence of dependencies for handlers for this
                collector.
        """
        self.collector.include_collector(collector, dependencies=dependencies)

    def command_for(self, *args: Any) -> str:
        """Find handler and build a command string using passed body query_params.

        Arguments:
            args: sequence of elements where first element should be name of handler.

        Returns:
            Command string.
        """
        return self.collector.command_for(*args)

    def handler_for(self, name: str) -> Handler:
        """Find handler in handlers of this bot.

        Find registered handler using using [botx.collector.Collector.handler_for] of
        inner collector.

        Arguments:
            name: name of handler that should be found.

        Returns:
            Handler that was found by name.
        """
        return self.collector.handler_for(name)



"""Definition for base class that is responsible for request to BotX API."""
from __future__ import annotations

import typing
from abc import ABC, abstractmethod

from httpx import URL, Response
from pydantic import BaseConfig, BaseModel, Extra
from pydantic.generics import GenericModel

from botx.clients.methods.request_wrapper import HTTPRequest, PrimitiveDataType
from botx.models.enums import Statuses

try:
    from typing import Literal  # noqa: WPS433
except ImportError:
    from typing_extensions import (  # type: ignore  # noqa: WPS433, WPS440, F401
        Literal,
    )

PRIMITIVES_FOR_QUERY = (str, int, float, bool, type(None))

ResponseT = typing.TypeVar("ResponseT")
SyncErrorHandler = typing.Callable[["BotXMethod", Response], typing.NoReturn]
AsyncErrorHandler = typing.Callable[
    ["BotXMethod", Response], typing.Awaitable[typing.NoReturn],
]
ErrorHandler = typing.Union[SyncErrorHandler, AsyncErrorHandler]
ErrorHandlersInMethod = typing.Union[typing.Sequence[ErrorHandler], ErrorHandler]


def _convert_query_to_primitives(
    query_params: typing.Mapping[str, typing.Any],
) -> typing.Dict[str, PrimitiveDataType]:
    converted_params = {}
    for param_key, param_value in query_params.items():
        if isinstance(param_value, PRIMITIVES_FOR_QUERY):
            converted_params[param_key] = param_value
        else:
            converted_params[param_key] = str(param_value)

    return converted_params


class APIResponse(GenericModel, typing.Generic[ResponseT]):
    """Model for successful response from BotX API."""

    #: status of requested operation response.
    status: Literal[Statuses.ok] = Statuses.ok

    #: generic response shape.
    result: ResponseT


class APIErrorResponse(GenericModel, typing.Generic[ResponseT]):
    """Model for error response from BotX API."""

    #: status of requested operation response.
    status: Literal[Statuses.error] = Statuses.error

    #: reason why operation failed
    reason: str

    #: errors from API.
    errors: typing.List[str]

    #: additional payload with more data about error.
    error_data: ResponseT


class AbstractBotXMethod(ABC, typing.Generic[ResponseT]):
    """Abstract base class for BotX request."""

    @property
    @abstractmethod
    def __url__(self) -> str:
        """Path for method in BotX API."""

    @property
    @abstractmethod
    def __method__(self) -> str:
        """HTTP method used for method."""

    @property
    @abstractmethod
    def __returning__(self) -> typing.Type[typing.Any]:
        """Shape returned from method that can be parsed by pydantic."""

    @property
    def __errors_handlers__(self) -> typing.Mapping[int, ErrorHandlersInMethod]:
        """Error handlers for responses from BotX API by status code and handler."""
        return typing.cast(typing.Mapping[int, ErrorHandlersInMethod], {})

    @property
    def __result_extractor__(
        self,
    ) -> typing.Optional[typing.Callable[[BotXMethod, typing.Any], ResponseT]]:
        """Extractor for response shape from BotX API."""
        return None  # noqa: WPS324


CREDENTIALS_FIELDS = ("token", "host", "scheme")


class BaseBotXMethod(AbstractBotXMethod[ResponseT], ABC):
    """Base logic that is responsible for configuration and shortcuts for fields."""

    #: host where request should be sent.
    host: str = ""

    #: token for request.
    token: str = ""

    #: HTTP scheme for request.
    scheme: str = "https"

    @property
    def url(self) -> str:
        """Full URL for request."""
        base_url = "{scheme}://{host}".format(scheme=self.scheme, host=self.host)
        return str(URL(base_url).join(self.__url__))

    @property
    def http_method(self) -> str:
        """HTTP method for request."""
        return self.__method__

    @property
    def headers(self) -> typing.Dict[str, str]:
        """Headers that should be used in request."""
        return {"Content-Type": "application/json"}

    @property
    def query_params(self) -> typing.Dict[str, PrimitiveDataType]:
        """Query string query_params for request."""
        return {}

    @property
    def returning(self) -> typing.Type[typing.Any]:
        """Shape returned from method that can be parsed by pydantic."""
        return self.__returning__

    @property
    def error_handlers(self) -> typing.Mapping[int, ErrorHandlersInMethod]:
        """Error handlers for responses from BotX API by status code and handler."""
        return self.__errors_handlers__

    @property
    def result_extractor(
        self,
    ) -> typing.Optional[typing.Callable[[BotXMethod, typing.Any], ResponseT]]:
        """Extractor for response shape from BotX API."""
        return self.__result_extractor__


class BotXMethod(BaseBotXMethod[ResponseT], BaseModel, ABC):
    """Method for BotX API that should be extended by actual implementation."""

    class Config(BaseConfig):
        extra = Extra.allow
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        orm_mode = True

    def configure(self, *, host: str, token: str, scheme: str = "https") -> None:
        """Configure request with credentials and transport related stuff.

        Arguments:
            host: host where request should be sent.
            token: token for request.
            scheme: HTTP scheme for request.
        """
        self.token = token
        self.host = host
        self.scheme = scheme

    def encode(self) -> typing.Optional[str]:
        """Encode request data body for request.

        Returns:
            Encoded data for request.
        """
        return self.json(
            by_alias=True, exclude=set(CREDENTIALS_FIELDS), exclude_none=True,
        )

    def build_http_request(self) -> HTTPRequest:
        """Build HTTP request that can be used by clients for making real requests.

        Returns:
            Built HTTP request.
        """
        request_params = self.query_params
        request_data = self.encode()

        if self.__method__ == "GET":
            if request_data:
                request_data = None

            if not request_params:
                request_params = _convert_query_to_primitives(
                    self.dict(exclude=set(CREDENTIALS_FIELDS)),
                )
                request_data = None

        return HTTPRequest(
            method=self.__method__,
            url=self.url,
            headers=self.headers,
            query_params=dict(request_params),
            request_data=request_data,
        )


class AuthorizedBotXMethod(BotXMethod[ResponseT], ABC):
    """Method for BotX API that adds authorization token."""

    @property
    def headers(self) -> typing.Dict[str, str]:
        """Headers that should be used in request."""
        headers = super().headers
        headers["Authorization"] = "Bearer {token}".format(token=self.token)
        return headers



"""Mixin that defines handler decorator."""

from typing import Callable, Optional, Sequence, Union

from botx.collecting.collectors.collector import Collector
from botx.dependencies.models import Depends


class AddHandlerMixin:
    """Mixin that defines handler decorator."""

    collector: Collector

    def add_handler(  # noqa: WPS211
        self,
        handler: Callable,
        *,
        body: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        full_description: Optional[str] = None,
        include_in_status: Union[bool, Callable] = True,
        dependencies: Optional[Sequence[Depends]] = None,
    ) -> None:
        """Create new handler from passed arguments and store it inside.

        !!! info
            If `include_in_status` is a function, then `body` argument will be checked
            for matching public commands style, like `/command`.

        Arguments:
            handler: callable that will be used for executing handler.
            body: body template that will trigger this handler.
            name: optional name for handler that will be used in generating body.
            description: description for command that will be shown in bot's menu.
            full_description: full description that can be used for example in `/help`
                command.
            include_in_status: should this handler be shown in bot's menu, can be
                callable function with no arguments *(for now)*.
            dependencies: sequence of dependencies that should be executed before
                handler.
        """
        self.collector.add_handler(
            body=body,
            handler=handler,
            name=name,
            description=description,
            full_description=full_description,
            include_in_status=include_in_status,
            dependencies=dependencies,
        )



"""Base for testing client for bots."""
from __future__ import annotations

import asyncio
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from typing import Any, Dict, Generator, List, Optional, Tuple, Type

import httpx

from botx.bots.bots import Bot
from botx.clients.methods.base import BotXMethod
from botx.middlewares.exceptions import ExceptionMiddleware
from botx.models.messages.incoming_message import IncomingMessage
from botx.models.messages.message import Message
from botx.testing.botx_mock.asgi.application import get_botx_asgi_api
from botx.testing.botx_mock.wsgi.application import get_botx_wsgi_api
from botx.testing.typing import APIMessage, APIRequest


class _ExceptionMiddleware(ExceptionMiddleware):
    """Replacement of built-in ExceptionMiddleware that will raise errors."""

    async def _handle_error_in_handler(self, exc: Exception, message: Message) -> None:
        handler = self._lookup_handler_for_exception(exc)

        if handler is None:
            raise exc

        await super()._handle_error_in_handler(exc, message)


ErrorsOverrides = Dict[Type[BotXMethod], Tuple[int, Any]]


class BaseTestClient:
    """Base for testing client for bots."""

    def __init__(
        self,
        bot: Bot,
        errors: Optional[ErrorsOverrides] = None,
        suppress_errors: bool = False,
    ) -> None:
        """Init client with required query_params.

        Arguments:
            bot: bot that should be tested.
            errors: errors that should be raised from methods calls.
            suppress_errors: if True then don't raise raise errors from handlers.
        """
        self.bot: Bot = bot
        self._original_http_client = bot.client.http_client
        self._original_sync_http_client = bot.sync_client.http_client
        self._error_middleware: Optional[ExceptionMiddleware] = None
        self._messages: List[APIMessage] = []
        self._requests: List[APIRequest] = []
        self._errors = errors or {}
        self._suppress_errors = suppress_errors

    def __enter__(self) -> BaseTestClient:
        """Mock original HTTP clients."""
        is_error_middleware = isinstance(
            self.bot.exception_middleware, ExceptionMiddleware,
        )
        if not self._suppress_errors and is_error_middleware:
            self._error_middleware = self.bot.exception_middleware
            self.bot.exception_middleware = _ExceptionMiddleware(
                self.bot.exception_middleware.executor,
            )
            self.bot.exception_middleware._exception_handlers = (  # noqa: WPS437
                self._error_middleware._exception_handlers  # noqa: WPS437
            )

        self.bot.client.http_client = httpx.AsyncClient(
            app=get_botx_asgi_api(self._messages, self._requests, self._errors),
        )
        self.bot.sync_client.http_client = httpx.Client(
            app=get_botx_wsgi_api(self._messages, self._requests, self._errors),
        )

        return self

    def __exit__(self, *_: Any) -> None:
        """Restore original HTTP client and clear storage."""
        if self._error_middleware is not None:
            self.bot.exception_middleware = self._error_middleware

        ThreadPoolExecutor().submit(
            asyncio.run, self.bot.client.http_client.aclose(),
        ).result()
        self.bot.client.http_client = self._original_http_client
        self.bot.sync_client.http_client = self._original_sync_http_client
        self._messages = []

    @contextmanager
    def error_client(
        self, errors: Dict[Type[BotXMethod], Tuple[int, Any]],
    ) -> Generator[BaseTestClient, None, None]:
        """Enter into new test client that adds error responses to mocks.

        Arguments:
            errors: overrides for errors in context.

        Yields:
            New client with overridden errors.
        """
        override_errors = {**self._errors, **errors}
        with self.__class__(self.bot, override_errors, self._suppress_errors) as client:
            yield client

    async def send_command(self, message: IncomingMessage, sync: bool = True) -> None:
        """Send command message to bot.

        Arguments:
            message: message with command for bot.
            sync: if is `True` then wait while command is full executed.
        """
        await self.bot.execute_command(message.dict())

        if sync:
            await self.bot.wait_current_handlers()



"""Definition for markup attached to sent message."""

from typing import List, Optional, Type, TypeVar

from pydantic import BaseModel

from botx.models.buttons import BubbleElement, Button, ButtonOptions, KeyboardElement

TUIElement = TypeVar("TUIElement", bound=Button)


class MessageMarkup(BaseModel):
    """Collection for bubbles and keyboard with some helper methods."""

    #: bubbles that will be attached to message.
    bubbles: List[List[BubbleElement]] = []

    #: keyboard elements that will be attached to message.
    keyboard: List[List[KeyboardElement]] = []

    def add_bubble(  # noqa: WPS211
        self,
        command: str,
        label: Optional[str] = None,
        data: Optional[dict] = None,  # noqa: WPS110
        options: Optional[ButtonOptions] = None,
        *,
        new_row: bool = True,
    ) -> None:
        """Add new bubble button to markup.

        Arguments:
            command: command that will be triggered on bubble click.
            label: label that will be shown on bubble.
            data: payload that will be attached to bubble.
            options: add special effects to bubble.
            new_row: place bubble on new row or on current.
        """
        self._add_ui_element(
            ui_cls=BubbleElement,
            ui_array=self.bubbles,
            command=command,
            label=label,
            data=data,
            opts=options,
            new_row=new_row,
        )

    def add_bubble_element(
        self, element: BubbleElement, *, new_row: bool = True,
    ) -> None:
        """Add new button to markup from existing element.

        Arguments:
            element: existed bubble element.
            new_row: place bubble on new row or on current.
        """
        self._add_ui_element(
            ui_cls=BubbleElement,
            ui_array=self.bubbles,
            command=element.command,
            label=element.label,
            data=element.data,
            opts=element.opts,
            new_row=new_row,
        )

    def add_keyboard_button(  # noqa: WPS211
        self,
        command: str,
        label: Optional[str] = None,
        data: Optional[dict] = None,  # noqa: WPS110
        options: Optional[ButtonOptions] = None,
        *,
        new_row: bool = True,
    ) -> None:
        """Add new keyboard button to markup.

        Arguments:
            command: command that will be triggered on keyboard click.
            label: label that will be shown on keyboard button.
            data: payload that will be attached to keyboard.
            options: add special effects to keyboard button.
            new_row: place keyboard on new row or on current.
        """
        self._add_ui_element(
            ui_cls=KeyboardElement,
            ui_array=self.keyboard,
            command=command,
            label=label,
            data=data,
            opts=options,
            new_row=new_row,
        )

    def add_keyboard_button_element(
        self, element: KeyboardElement, *, new_row: bool = True,
    ) -> None:
        """Add new keyboard button to markup from existing element.

        Arguments:
            element: existed keyboard button element.
            new_row: place keyboard button on new row or on current.
        """
        self._add_ui_element(
            ui_cls=KeyboardElement,
            ui_array=self.keyboard,
            command=element.command,
            label=element.label,
            data=element.data,
            opts=element.opts,
            new_row=new_row,
        )

    def _add_ui_element(  # noqa: WPS211
        self,
        ui_cls: Type[TUIElement],
        ui_array: List[List[TUIElement]],
        command: str,
        label: Optional[str] = None,
        data: Optional[dict] = None,  # noqa: WPS110
        opts: Optional[ButtonOptions] = None,
        new_row: bool = True,
    ) -> None:
        """Add new button to bubble or keyboard arrays.

        Arguments:
            ui_cls: UIElement instance that should be added to array.
            ui_array: storage for ui elements.
            command: command that will be triggered on ui element click.
            label: label that will be shown on ui element.
            data: payload that will be attached to ui element.
            opts: add special effects ui element.
            new_row: place ui element on new row or on current.
        """
        element = ui_cls(
            command=command,
            label=label,
            data=(data or {}),
            opts=(opts or ButtonOptions()),
        )

        if new_row:
            ui_array.append([element])
            return

        if not ui_array:
            ui_array.append([])

        ui_array[-1].append(element)



"""Method for promoting users to admins."""
from typing import List
from uuid import UUID

import httpx

from botx.clients.methods.base import AuthorizedBotXMethod
from botx.clients.methods.errors import (
    bot_is_not_admin,
    chat_is_not_modifiable,
    chat_not_found,
)


class AddAdminRole(AuthorizedBotXMethod[bool]):
    """Method for promoting users to chat admins."""

    __url__ = "/api/v3/botx/chats/add_admin"
    __method__ = "POST"
    __returning__ = bool
    __errors_handlers__ = {
        httpx.codes.FORBIDDEN: (
            bot_is_not_admin.handle_error,
            chat_is_not_modifiable.handle_error,
        ),
        httpx.codes.NOT_FOUND: (chat_not_found.handle_error,),
    }

    #: ID of chat where action should be performed.
    group_chat_id: UUID

    #: IDs of users that should be promoted to admins.
    user_huids: List[UUID]



"""Definition of file that can be included in incoming message or in sending result."""

import base64
import mimetypes
import pathlib
from io import BytesIO
from typing import AnyStr, BinaryIO, Optional, TextIO, Union

from pydantic import BaseModel, validator

#: file extensions that can be proceed by BotX API.
BOTX_API_ACCEPTED_EXTENSIONS = (
    # image extensions
    ".jpg",
    ".jpeg",
    ".gif",
    ".png",
    # document extensions
    ".doc",
    ".docx",
    ".xls",
    ".xlsx",
    ".txt",
    ".pdf",
    ".html",
    ".json",
    ".gz",
    ".tgz",
    ".zip",
    ".rar",
    ".ppt",
    ".pptx",
)


class File(BaseModel):
    """Object that represents file in RFC 2397 format."""

    #: name of file.
    file_name: str

    #: file content in RFC 2397 format.
    data: str  # noqa: WPS110

    #: text under file.
    caption: Optional[str] = None

    @validator("file_name", always=True)
    def check_file_extension(cls, name: str) -> str:  # noqa: N805
        """Check that file extension can be handled by BotX API.

        Arguments:
            cls: this class.
            name: file name which will be checked for matching extensions.

        Returns:
            Passed name if matching was successful.

        Raises:
            ValueError: raised if extension is not supported.
        """
        extensions_check = (
            name.lower().endswith(extension)
            for extension in BOTX_API_ACCEPTED_EXTENSIONS
        )

        if not any(extensions_check):
            raise ValueError(
                "file {0} has an extensions that is not supported by BotX API".format(
                    name,
                ),
            )

        return name

    @classmethod
    def from_file(
        cls, file: Union[TextIO, BinaryIO], filename: Optional[str] = None,
    ) -> "File":
        """Convert file-like object into BotX API compatible file.

        Arguments:
            file: file-like object that will be used for creating file.
            filename: name that will be used for file, if was not passed, then will be
                retrieved from `file` `.name` property.

        Returns:
            Built file object.
        """
        filename = filename or pathlib.Path(file.name).name
        file_data = file.read()

        if isinstance(file_data, str):
            file_data = file_data.encode()

        encoded_data = base64.b64encode(file_data).decode()
        media_type = mimetypes.guess_type(filename)[0] or "text/plain"
        return cls(
            file_name=filename,
            data="data:{0};base64,{1}".format(media_type, encoded_data),
        )

    @classmethod
    def from_string(cls, data_of_file: AnyStr, filename: str) -> "File":
        """Build file from bytes or string passed to method in `data` with `filename` as name.

        Arguments:
            data_of_file: string or bytes that will be used for creating file.
            filename: name for new file.

        Returns:
            Built file object.
        """
        if isinstance(data_of_file, str):
            file_data = data_of_file.encode()
        else:
            file_data = data_of_file
        file = BytesIO(file_data)
        file.name = filename
        return cls.from_file(file)

    @property
    def file(self) -> BinaryIO:
        """Return file data in file-like object that will return bytes."""
        bytes_file = BytesIO(self.data_in_bytes)
        bytes_file.name = self.file_name
        return bytes_file

    @property
    def data_in_bytes(self) -> bytes:
        """Return decoded file data in bytes."""
        return base64.b64decode(self.data_in_base64)

    @property
    def data_in_base64(self) -> str:
        """Return file data in base64 encoded string."""
        return self.data.split(",", 1)[1]

    @property
    def media_type(self) -> str:
        """Return media type of file."""
        return mimetypes.guess_type(self.data)[0] or "text/plain"



"""Definition of different schemas for system events."""

from types import MappingProxyType
from typing import List, Mapping, Type
from uuid import UUID

from pydantic import BaseModel

from botx.models.enums import ChatTypes, SystemEvents
from botx.models.users import UserInChatCreated


class ChatCreatedEvent(BaseModel):
    """Shape for `system:chat_created` event data."""

    #: chat id from which event received.
    group_chat_id: UUID

    #: type of chat.
    chat_type: ChatTypes

    #: chat name.
    name: str

    #: HUID of user that created chat.
    creator: UUID

    #: list of users that are members of chat.
    members: List[UserInChatCreated]


class AddedToChatEvent(BaseModel):
    """Shape for `system:added_to_chat` event data."""

    #: members added to chat.
    added_members: List[UUID]


# dict for validating shape for different events
EVENTS_SHAPE_MAP: Mapping[SystemEvents, Type[BaseModel]] = MappingProxyType(
    {
        SystemEvents.chat_created: ChatCreatedEvent,
        SystemEvents.added_to_chat: AddedToChatEvent,
    },
)



"""Method for sending notification into single chat."""
from typing import Optional
from uuid import UUID

from pydantic import Field

from botx.clients.methods.base import AuthorizedBotXMethod
from botx.clients.methods.extractors import extract_generated_sync_id
from botx.clients.types.message_payload import ResultPayload
from botx.clients.types.options import ResultOptions
from botx.clients.types.response_results import PushResult
from botx.models.files import File
from botx.models.typing import AvailableRecipients


class NotificationDirect(AuthorizedBotXMethod[UUID]):
    """Method for sending notification into single chat."""

    __url__ = "/api/v3/botx/notification/callback/direct"
    __method__ = "POST"
    __returning__ = PushResult
    __result_extractor__ = extract_generated_sync_id

    #: ID of chat for new notification.
    group_chat_id: UUID

    #: custom ID for message.
    event_sync_id: Optional[UUID] = None

    #: HUIDs of users that should receive notifications.
    recipients: AvailableRecipients = "all"

    #: data for build message: body, markup, mentions.
    result: ResultPayload = Field(..., alias="notification")

    #: attached file for message.
    file: Optional[File] = None

    #: extra options for message.
    opts: ResultOptions = ResultOptions()



"""Message that is sent from bot."""
from typing import Any, BinaryIO, Dict, List, Optional, TextIO, Union, cast
from uuid import UUID

from botx.models.buttons import ButtonOptions
from botx.models.enums import MentionTypes
from botx.models.files import File
from botx.models.mentions import ChatMention, Mention, UserMention
from botx.models.messages.message import Message
from botx.models.messages.sending.credentials import SendingCredentials
from botx.models.messages.sending.markup import MessageMarkup
from botx.models.messages.sending.options import MessageOptions, NotificationOptions
from botx.models.messages.sending.payload import MessagePayload
from botx.models.typing import AvailableRecipients, BubbleMarkup, KeyboardMarkup

ARGUMENTS_DUPLICATION_ERROR = (
    "{0} can not be passed along with manual validated_values for it"
)


# currently I have no idea how to clean this.
class SendingMessage:  # noqa: WPS214
    """Message that will be sent by bot."""

    def __init__(  # noqa: WPS211
        self,
        *,
        text: str = "",
        bot_id: Optional[UUID] = None,
        host: Optional[str] = None,
        sync_id: Optional[UUID] = None,
        chat_id: Optional[UUID] = None,
        message_id: Optional[UUID] = None,
        recipients: Optional[AvailableRecipients] = None,
        mentions: Optional[List[Mention]] = None,
        bubbles: Optional[BubbleMarkup] = None,
        keyboard: Optional[KeyboardMarkup] = None,
        notification_options: Optional[NotificationOptions] = None,
        file: Optional[File] = None,
        credentials: Optional[SendingCredentials] = None,
        options: Optional[MessageOptions] = None,
        markup: Optional[MessageMarkup] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Init message with required attributes.

        !!! info
            You should pass at least already built credentials or bot_id, host and
            one of sync_id, chat_id or chat_ids for message.
        !!! info
            You can not pass markup along with bubbles or keyboards. You can merge them
            manual before or after building message.
        !!! info
            You can not pass options along with any of recipients, mentions or
            notification_options. You can merge them manual before or after building
            message.

        Arguments:
            text: text for message.
            file: file that will be attached to message.
            bot_id: bot id.
            host: host for message.
            sync_id: message event id.
            chat_id: chat id.
            message_id: custom id of new message.
            credentials: message credentials.
            bubbles: bubbles that will be attached to message.
            keyboard: keyboard elements that will be attached to message.
            markup: message markup.
            recipients: recipients for message.
            mentions: mentions that will be attached to message.
            notification_options: configuration for notifications for message.
            options: message options.
            metadata: message metadata.
        """
        self.credentials: SendingCredentials = _build_credentials(
            bot_id=bot_id,
            host=host,
            sync_id=sync_id,
            message_id=message_id,
            chat_id=chat_id,
            credentials=(credentials.copy() if credentials else credentials),
        )

        self.payload: MessagePayload = MessagePayload(
            text=text,
            metadata=metadata or {},
            file=file,
            markup=_build_markup(bubbles=bubbles, keyboard=keyboard, markup=markup),
            options=_build_options(
                recipients=recipients,
                mentions=mentions,
                notification_options=notification_options,
                options=options,
            ),
        )

    @classmethod
    def from_message(
        cls, *, text: str = "", file: Optional[File] = None, message: Message,
    ) -> "SendingMessage":
        """Build message for sending from incoming message.

        Arguments:
            text: text for message.
            file: file attached to message.
            message: incoming message.

        Returns:
            Built message.
        """
        return cls(
            text=text,
            file=file,
            sync_id=message.sync_id,
            chat_id=message.group_chat_id,
            bot_id=message.bot_id,
            host=message.host,
        )

    @property
    def text(self) -> str:
        """Text in message."""
        return self.payload.text

    @text.setter
    def text(self, text: str) -> None:
        """Text in message."""
        self.payload.text = text

    @property
    def metadata(self) -> Dict[str, Any]:
        """Metadata in message."""
        return self.payload.metadata

    @metadata.setter
    def metadata(self, metadata: Dict[str, Any]) -> None:
        self.payload.metadata = metadata

    @property
    def file(self) -> Optional[File]:
        """File attached to message."""
        return self.payload.file

    @file.setter
    def file(self, file: File) -> None:
        """File attached to message."""
        self.payload.file = file

    @property
    def markup(self) -> MessageMarkup:
        """Message markup."""
        return self.payload.markup

    @markup.setter
    def markup(self, markup: MessageMarkup) -> None:
        """Message markup."""
        self.payload.markup = markup

    @property
    def options(self) -> MessageOptions:
        """Message options."""
        return self.payload.options

    @options.setter
    def options(self, options: MessageOptions) -> None:
        """Message options."""
        self.payload.options = options

    @property
    def sync_id(self) -> Optional[UUID]:
        """Event id on which message should answer."""
        return self.credentials.sync_id

    @sync_id.setter
    def sync_id(self, sync_id: UUID) -> None:
        """Event id on which message should answer."""
        self.credentials.sync_id = sync_id

    @property
    def chat_id(self) -> Optional[UUID]:
        """Chat id in which message should be sent."""
        return self.credentials.chat_id

    @chat_id.setter
    def chat_id(self, chat_id: UUID) -> None:
        """Chat id in which message should be sent."""
        self.credentials.chat_id = chat_id

    @property
    def bot_id(self) -> UUID:
        """Bot id that handles message."""
        return cast(UUID, self.credentials.bot_id)

    @bot_id.setter
    def bot_id(self, bot_id: UUID) -> None:
        """Bot id that handles message."""
        self.credentials.bot_id = bot_id

    @property
    def host(self) -> str:
        """Host where BotX API places."""
        return cast(str, self.credentials.host)

    @host.setter
    def host(self, host: str) -> None:
        """Host where BotX API places."""
        self.credentials.host = host

    def add_file(
        self, file: Union[TextIO, BinaryIO, File], filename: Optional[str] = None,
    ) -> None:
        """Attach file to message.

        Arguments:
            file: file that should be attached to the message.
            filename: name for file that will be used if if can not be retrieved from
                file.
        """
        if isinstance(file, File):
            file.file_name = filename or file.file_name
            self.payload.file = file
        else:
            self.payload.file = File.from_file(file, filename=filename)

    def mention_user(self, user_huid: UUID, name: Optional[str] = None) -> None:
        """Mention user in message.

        Arguments:
            user_huid: id of user that should be mentioned.
            name: name that will be shown.
        """
        self.payload.options.mentions.append(
            Mention(mention_data=UserMention(user_huid=user_huid, name=name)),
        )

    def mention_contact(self, user_huid: UUID, name: Optional[str] = None) -> None:
        """Mention contact in message.

        Arguments:
            user_huid: id of user that should be mentioned.
            name: name that will be shown.
        """
        self.payload.options.mentions.append(
            Mention(
                mention_data=UserMention(user_huid=user_huid, name=name),
                mention_type=MentionTypes.contact,
            ),
        )

    def mention_chat(self, group_chat_id: UUID, name: Optional[str] = None) -> None:
        """Mention chat in message.

        Arguments:
            group_chat_id: id of chat that should be mentioned.
            name: name that will be shown.
        """
        self.payload.options.mentions.append(
            Mention(
                mention_data=ChatMention(group_chat_id=group_chat_id, name=name),
                mention_type=MentionTypes.chat,
            ),
        )

    def add_recipient(self, recipient: UUID) -> None:
        """Add new user that will receive message.

        Arguments:
            recipient: recipient for message.
        """
        if self.payload.options.recipients == "all":
            self.payload.options.recipients = []

        self.payload.options.recipients.append(recipient)

    def add_recipients(self, recipients: List[UUID]) -> None:
        """Add list of recipients that should receive message.

        Arguments:
            recipients: recipients for message.
        """
        if self.payload.options.recipients == "all":
            self.payload.options.recipients = []

        self.payload.options.recipients.extend(recipients)

    def add_bubble(  # noqa: WPS211
        self,
        command: str,
        label: Optional[str] = None,
        data: Optional[dict] = None,  # noqa: WPS110
        options: Optional[ButtonOptions] = None,
        *,
        new_row: bool = True,
    ) -> None:
        """Add new bubble button to message markup.

        Arguments:
            command: command that will be triggered on bubble click.
            label: label that will be shown on bubble.
            data: payload that will be attached to bubble.
            options: add special effects to bubble.
            new_row: place bubble on new row or on current.
        """
        self.payload.markup.add_bubble(command, label, data, options, new_row=new_row)

    def add_keyboard_button(  # noqa: WPS211
        self,
        command: str,
        label: Optional[str] = None,
        data: Optional[dict] = None,  # noqa: WPS110
        options: Optional[ButtonOptions] = None,
        *,
        new_row: bool = True,
    ) -> None:
        """Add new keyboard button to message markup.

        Arguments:
            command: command that will be triggered on keyboard click.
            label: label that will be shown on keyboard button.
            data: payload that will be attached to keyboard.
            options: add special effects to keyboard button.
            new_row: place keyboard on new row or on current.
        """
        self.payload.markup.add_keyboard_button(
            command, label, data, options, new_row=new_row,
        )

    def show_notification(self, show: bool) -> None:
        """Show notification about message.

        Arguments:
            show: show notification about message.
        """
        self.payload.options.notifications.send = show

    def force_notification(self, force: bool) -> None:
        """Break mute on bot messages.

        Arguments:
            force: break mute on bot messages.
        """
        self.payload.options.notifications.force_dnd = force


def _build_credentials(  # noqa: WPS211
    bot_id: Optional[UUID] = None,
    host: Optional[str] = None,
    sync_id: Optional[UUID] = None,
    message_id: Optional[UUID] = None,
    chat_id: Optional[UUID] = None,
    credentials: Optional[SendingCredentials] = None,
) -> SendingCredentials:
    """Build credentials for message.

    Arguments:
        bot_id: bot id.
        host: host for message.
        sync_id: message event id.
        message_id: id of new message.
        chat_id: chat id.
        credentials: message credentials.

    Returns:
        Credentials for message.

    Raises:
        AssertionError: raised if credentials were passed with separate parameters.
    """
    if bot_id and host:
        if credentials is not None:
            raise AssertionError(
                ARGUMENTS_DUPLICATION_ERROR.format("MessageCredentials"),
            )

        return SendingCredentials(
            bot_id=bot_id,
            host=host,
            sync_id=sync_id,
            chat_id=chat_id,
            message_id=message_id,
        )

    if credentials is None:
        raise AssertionError(
            "MessageCredentials or manual validated_values should be passed",
        )

    if credentials.message_id is None:
        credentials.message_id = message_id

    return credentials


def _build_markup(
    bubbles: Optional[BubbleMarkup] = None,
    keyboard: Optional[KeyboardMarkup] = None,
    markup: Optional[MessageMarkup] = None,
) -> MessageMarkup:
    """Build markup for message.

    Arguments:
        bubbles: bubbles that will be attached to message.
        keyboard: keyboard elements that will be attached to message.
        markup: message markup.

    Returns:
        Markup for message.

    Raises:
        AssertionError: raised if markup were passed with separate parameters.
    """
    if bubbles is not None or keyboard is not None:
        if markup is not None:
            raise AssertionError(
                "Markup can not be passed along with bubbles or keyboard elements",
            )
        return MessageMarkup(bubbles=bubbles or [], keyboard=keyboard or [])

    return markup or MessageMarkup()


def _build_options(
    recipients: Optional[AvailableRecipients] = None,
    mentions: Optional[List[Mention]] = None,
    notification_options: Optional[NotificationOptions] = None,
    options: Optional[MessageOptions] = None,
) -> MessageOptions:
    """Build options for message.

    Arguments:
        recipients: recipients for message.
        mentions: mentions that will be attached to message.
        notification_options: configuration for notifications for message.
        options: message options.

    Returns:
        Options for message.

    Raises:
        AssertionError: raised if options were passed with separate parameters.
    """
    if mentions or recipients or notification_options:
        if options is not None:
            raise AssertionError(ARGUMENTS_DUPLICATION_ERROR.format("MessageOptions"))
        return MessageOptions(
            recipients=recipients or "all",
            mentions=mentions or [],
            notifications=notification_options or NotificationOptions(),
        )

    return options or MessageOptions()



"""Endpoints for notification resource."""

from starlette.requests import Request
from starlette.responses import Response

from botx.clients.methods.base import APIResponse
from botx.clients.methods.v3.notification.direct_notification import NotificationDirect
from botx.clients.methods.v3.notification.notification import Notification
from botx.testing.botx_mock.asgi.messages import add_message_to_collection
from botx.testing.botx_mock.asgi.responses import (
    PydanticResponse,
    generate_push_response,
)
from botx.testing.botx_mock.binders import bind_implementation_to_method


@bind_implementation_to_method(Notification)
async def post_notification(request: Request) -> Response:
    """Handle pushed notification request.

    Arguments:
        request: HTTP request from Starlette.

    Returns:
        Response with sync_id of pushed message.
    """
    payload = Notification.parse_obj(await request.json())
    add_message_to_collection(request, payload)
    return PydanticResponse(APIResponse[str](result="notification_pushed"))


@bind_implementation_to_method(NotificationDirect)
async def post_notification_direct(request: Request) -> Response:
    """Handle pushed notification request.

    Arguments:
        request: HTTP request from Starlette.

    Returns:
        Response with sync_id of pushed message.
    """
    payload = NotificationDirect.parse_obj(await request.json())
    add_message_to_collection(request, payload)
    return generate_push_response(payload)



"""Definition for mixin that defines helpers for sending message."""

from typing import BinaryIO, Optional, TextIO, Union
from uuid import UUID

from botx.models.files import File
from botx.models.messages.message import Message
from botx.models.messages.sending.credentials import SendingCredentials
from botx.models.messages.sending.markup import MessageMarkup
from botx.models.messages.sending.message import SendingMessage
from botx.models.messages.sending.options import MessageOptions
from botx.models.messages.sending.payload import MessagePayload, UpdatePayload

try:
    from typing import Protocol  # noqa: WPS433
except ImportError:
    from typing_extensions import (  # type: ignore  # noqa: WPS433, WPS440, F401
        Protocol,
    )


class ResultSendProtocol(Protocol):
    """Protocol for object that can create new or update message."""

    async def send_command_result(
        self, credentials: SendingCredentials, payload: MessagePayload,
    ) -> UUID:
        """Send command result."""

    async def send_direct_notification(
        self, credentials: SendingCredentials, payload: MessagePayload,
    ) -> UUID:
        """Send notification."""

    async def update_message(
        self, credentials: SendingCredentials, update: UpdatePayload,
    ) -> None:
        """Update existing message."""


class MessageSendProtocol(Protocol):
    """Protocol for object that can send complex message."""

    async def send(self, message: SendingMessage) -> UUID:
        """Send message."""


class SendingMixin:
    """Mixin that defines helpers for sending messages."""

    async def send_message(  # noqa: WPS211
        self: MessageSendProtocol,
        text: str,
        credentials: SendingCredentials,
        *,
        file: Optional[Union[BinaryIO, TextIO]] = None,
        markup: Optional[MessageMarkup] = None,
        options: Optional[MessageOptions] = None,
    ) -> UUID:
        """Send message as answer to command or notification to chat and get it id.

        Arguments:
            text: text that should be sent to client.
            credentials: credentials that are used for sending message.
            file: file that should be attached to message.
            markup: message markup that should be attached to message.
            options: extra options for message.

        Returns:
            `UUID` of sent event.
        """
        message = SendingMessage(
            text=text, markup=markup, options=options, credentials=credentials,
        )
        if file:
            message.add_file(file)

        return await self.send(message)

    async def send(
        self: ResultSendProtocol, message: SendingMessage, *, update: bool = False,
    ) -> UUID:
        """Send message as answer to command or notification to chat and get it id.

        Arguments:
            message: message that should be sent to chat.
            update: if True then check, that `message_id` was set in credentials and
                update existing message with this ID.

        Returns:
            `UUID` of sent event.
        """
        if message.credentials.message_id is not None and update:
            await self.update_message(
                message.credentials.copy(
                    update={"sync_id": message.credentials.message_id},
                ),
                UpdatePayload.from_sending_payload(message.payload),
            )
            return message.credentials.message_id

        if message.sync_id:
            return await self.send_command_result(message.credentials, message.payload)

        return await self.send_direct_notification(message.credentials, message.payload)

    async def answer_message(  # noqa: WPS211
        self: MessageSendProtocol,
        text: str,
        message: Message,
        *,
        file: Optional[Union[BinaryIO, TextIO, File]] = None,
        markup: Optional[MessageMarkup] = None,
        options: Optional[MessageOptions] = None,
    ) -> UUID:
        """Answer on incoming message and return id of new message..

        !!! warning
            This method should be used only in handlers.

        Arguments:
            text: text that should be sent in message.
            message: incoming message.
            file: file that can be attached to the message.
            markup: bubbles and keyboard that can be attached to the message.
            options: additional message options, like mentions or notifications
                configuration.

        Returns:
            `UUID` of sent event.
        """
        sending_message = SendingMessage(
            text=text, credentials=message.credentials, markup=markup, options=options,
        )
        if file:
            sending_message.add_file(file)

        return await self.send(sending_message)

    async def send_file(  # noqa: WPS211
        self: MessageSendProtocol,
        file: Union[TextIO, BinaryIO, File],
        credentials: SendingCredentials,
        filename: Optional[str] = None,
    ) -> UUID:
        """Send file in chat and return id of message.

        Arguments:
            file: file-like object that will be sent to chat.
            credentials: credentials of chat where file should be sent.
            filename: name for file that will be used if it can not be accessed from
                `file` argument.

        Returns:
            `UUID` of sent event.
        """
        message = SendingMessage(credentials=credentials)
        message.add_file(file, filename)
        return await self.send(message)
